#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import logging
import sys

import numpy as np
import sklearn
from scipy.stats import entropy
from sklearn.calibration import calibration_curve
from sklearn.metrics import brier_score_loss

from ludwig.constants import *
from ludwig.utils import visualization_utils
from ludwig.utils.data_utils import load_json, load_from_file
from ludwig.utils.print_utils import logging_level_registry


def learning_curves(training_statistics, field, model_names=None, **kwargs):
    if len(training_statistics) < 1:
        logging.error('No training_statistics provided')
        return

    training_statistics_per_model_name = [load_json(learning_stats_f)
                                          for learning_stats_f in
                                          training_statistics]

    fields_set = set()
    for ls in training_statistics_per_model_name:
        for _, values in ls.items():
            for key in values:
                fields_set.add(key)
    fields = [field] if field is not None and len(field) > 0 else fields_set

    metrics = [LOSS, ACCURACY, HITS_AT_K, EDIT_DISTANCE]
    for field in fields:
        for metric in metrics:
            if metric in training_statistics_per_model_name[0]['train'][field]:
                visualization_utils.lerning_curves_plot(
                    [learning_stats['train'][field][metric]
                     for learning_stats in training_statistics_per_model_name],
                    [learning_stats['validation'][field][metric]
                     for learning_stats in training_statistics_per_model_name],
                    metric, model_names,
                    title='Learning Curves {}'.format(field)
                )


def compare_performance(
        prediction_statistics,
        field, model_names=None,
        **kwargs
):
    if len(prediction_statistics) < 1:
        logging.error('No prediction_statistics provided')
        return

    prediction_statistics_per_model_name = [load_json(prediction_statistics_f)
                                            for prediction_statistics_f in
                                            prediction_statistics]

    fields_set = set()
    for ls in prediction_statistics_per_model_name:
        for key in ls:
            fields_set.add(key)
    fields = [field] if field is not None and len(field) > 0 else fields_set

    for field in fields:
        accuracies = []
        hits_at_ks = []
        edit_distances = []

        for prediction_statistics in prediction_statistics_per_model_name:
            if ACCURACY in prediction_statistics[field]:
                accuracies.append(prediction_statistics[field][ACCURACY])
            if HITS_AT_K in prediction_statistics[field]:
                hits_at_ks.append(prediction_statistics[field][HITS_AT_K])
            if EDIT_DISTANCE in prediction_statistics[field]:
                edit_distances.append(
                    prediction_statistics[field][EDIT_DISTANCE])

        measures = []
        measures_names = []
        if len(accuracies) > 0:
            measures.append(accuracies)
            measures_names.append(ACCURACY)
        if len(hits_at_ks) > 0:
            measures.append(hits_at_ks)
            measures_names.append(HITS_AT_K)
        if len(edit_distances) > 0:
            measures.append(edit_distances)
            measures_names.append(EDIT_DISTANCE)

        visualization_utils.compare_classifiers_plot(
            measures,
            measures_names,
            model_names,
            title='Performance comparison on {}'.format(field)
        )


def compare_classifiers_performance_from_prob(
        probabilities,
        ground_truth,
        field,
        top_n_classes,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    k = top_n_classes[0]

    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    accuracies = []
    hits_at_ks = []
    mrrs = []

    for i, prob in enumerate(probs):

        if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
            prob_limit = prob[:, :labels_limit + 1]
            prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
            prob = prob_limit

        prob = np.argsort(prob, axis=1)
        top1 = prob[:, -1]
        topk = prob[:, -k:]

        accuracies.append((gt == top1).sum() / len(gt))

        hits_at_k = 0
        for j in range(len(gt)):
            hits_at_k += np.in1d(gt[j], topk[i, :])
        hits_at_ks.append(np.asscalar(hits_at_k) / len(gt))

        mrr = 0
        for j in range(len(gt)):
            gt_pos_in_probs = prob[i, :] == gt[j]
            if np.any(gt_pos_in_probs):
                mrr += (1 / -(np.asscalar(np.argwhere(gt_pos_in_probs)) -
                              prob.shape[1]))
        mrrs.append(mrr / len(gt))

    visualization_utils.compare_classifiers_plot(
        [accuracies, hits_at_ks, mrrs],
        [ACCURACY, HITS_AT_K, 'mrr'],
        model_names
    )


def compare_classifiers_performance_from_pred(
        predictions,
        ground_truth,
        ground_truth_metadata,
        field,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(predictions) < 1:
        logging.error('No predictions provided')
        return

    metadata = load_json(ground_truth_metadata)

    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    preds = [load_from_file(preds_fn, dtype=str) for preds_fn in predictions]

    mapped_preds = []
    for pred in preds:
        mapped_preds.append([metadata[field]['str2idx'][val] for val in pred])
    preds = mapped_preds

    accuracies = []
    precisions = []
    recalls = []
    f1s = []

    for i, pred in enumerate(preds):
        accuracies.append(sklearn.metrics.accuracy_score(gt, pred))
        precisions.append(
            sklearn.metrics.precision_score(gt, pred, average='macro')
        )
        recalls.append(sklearn.metrics.recall_score(gt, pred, average='macro'))
        f1s.append(sklearn.metrics.f1_score(gt, pred, average='macro'))

    visualization_utils.compare_classifiers_plot(
        [accuracies, precisions, recalls, f1s],
        [ACCURACY, 'precision', 'recall', 'f1'],
        model_names
    )


def compare_classifiers_performance_subset(
        probabilities,
        ground_truth,
        field,
        top_n_classes,
        labels_limit,
        subset,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    k = top_n_classes[0]

    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    subset_indices = gt > 0
    gt_subset = gt
    if subset == 'ground_truth':
        subset_indices = gt < k
        gt_subset = gt[subset_indices]
        logging.info('Subset is {:.2f}% of the data'.format(
            len(gt_subset) / len(gt) * 100)
        )

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    accuracies = []
    hits_at_ks = []

    for i, prob in enumerate(probs):

        if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
            prob_limit = prob[:, :labels_limit + 1]
            prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
            prob = prob_limit

        if subset == PREDICTIONS:
            subset_indices = np.argmax(prob, axis=1) < k
            gt_subset = gt[subset_indices]
            logging.info(
                'Subset for model_name {} is {:.2f}% of the data'.format(
                    model_names[i] if model_names and i < len(
                        model_names) else i,
                    len(gt_subset) / len(gt) * 100
                )
            )
            model_names[i] = '{} ({:.2f}%)'.format(
                model_names[i] if model_names and i < len(model_names) else i,
                len(gt_subset) / len(gt) * 100
            )

        prob_subset = prob[subset_indices]

        prob_subset = np.argsort(prob_subset, axis=1)
        top1_subset = prob_subset[:, -1]
        top3_subset = prob_subset[:, -3:]

        accuracies.append(np.sum((gt_subset == top1_subset)) / len(gt_subset))

        hits_at_k = 0
        for j in range(len(gt_subset)):
            hits_at_k += np.in1d(gt_subset[j], top3_subset[i, :])
        hits_at_ks.append(np.asscalar(hits_at_k) / len(gt_subset))

    title = None
    if subset == 'ground_truth':
        title = 'Classifier performance on first {} class{} ({:.2f}%)'.format(
            k, 'es' if k > 1 else '', len(gt_subset) / len(gt) * 100
        )
    elif subset == PREDICTIONS:
        title = 'Classifier performance on first {} class{}'.format(
            k,
            'es' if k > 1 else ''
        )

    visualization_utils.compare_classifiers_plot(
        [accuracies, hits_at_ks],
        [ACCURACY, HITS_AT_K],
        model_names,
        title=title
    )


def compare_classifiers_performance_changing_k(
        probabilities,
        ground_truth,
        field,
        top_k,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    k = top_k

    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    hits_at_ks = []

    for i, prob in enumerate(probs):

        if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
            prob_limit = prob[:, :labels_limit + 1]
            prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
            prob = prob_limit

        prob = np.argsort(prob, axis=1)

        hits_at_k = [0.0] * k
        for g in range(len(gt)):
            for j in range(k):
                hits_at_k[j] += np.in1d(gt[g], prob[g, -j - 1:])
        hits_at_ks.append(np.array(hits_at_k) / len(gt))

    visualization_utils.compare_classifiers_line_plot(
        np.arange(1, k + 1),
        hits_at_ks, 'hits@k',
        model_names,
        title='Classifier comparison (hits@k)'
    )


def compare_classifiers_multiclass_multimetric(
        prediction_statistics,
        field,
        ground_truth_metadata,
        top_n_classes,
        model_names=None,
        **kwargs
):
    if len(prediction_statistics) < 1:
        logging.error('No prediction_statistics provided')
        return

    metadata = load_json(ground_truth_metadata)
    prediction_statistics_per_model_name = [load_json(prediction_statistics_f)
                                            for prediction_statistics_f in
                                            prediction_statistics]

    fields_set = set()
    for ls in prediction_statistics_per_model_name:
        for key in ls:
            fields_set.add(key)
    fields = [field] if field is not None and len(field) > 0 else fields_set

    for i, prediction_statistics in enumerate(
            prediction_statistics_per_model_name):
        for field in fields:
            model_name_name = (
                model_names[i]
                if model_names is not None and i < len(model_names)
                else ''
            )
            per_class_stats = prediction_statistics[field]['per_class_stats']
            precisions = []
            recalls = []
            f1_scores = []
            labels = []
            for _, class_name in sorted(
                    [(metadata[field]['str2idx'][key], key)
                     for key in per_class_stats.keys()],
                    key=lambda tup: tup[0]):
                class_stats = per_class_stats[class_name]
                precisions.append(class_stats['precision'])
                recalls.append(class_stats['recall'])
                f1_scores.append(class_stats['f1_score'])
                labels.append(class_name)
            for k in top_n_classes:
                k = min(k, len(precisions)) if k > 0 else len(precisions)
                ps = precisions[0:k]
                rs = recalls[0:k]
                fs = f1_scores[0:k]
                ls = labels[0:k]
                visualization_utils.compare_classifiers_multiclass_multimetric_plot(
                    [ps, rs, fs],
                    ['precision', 'recall', 'f1 score'],
                    labels=ls,
                    title='{} Multiclass Precision / Recall / '
                          'F1 Score top {} {}'.format(model_name_name, k, field)
                )

            p_np = np.nan_to_num(np.array(precisions, dtype=np.float32))
            r_np = np.nan_to_num(np.array(recalls, dtype=np.float32))
            f1_np = np.nan_to_num(np.array(f1_scores, dtype=np.float32))
            labels_np = np.nan_to_num(np.array(labels))

            sorted_indices = f1_np.argsort()
            higher_f1s = sorted_indices[-k:][::-1]
            visualization_utils.compare_classifiers_multiclass_multimetric_plot(
                [p_np[higher_f1s],
                 r_np[higher_f1s],
                 f1_np[higher_f1s]],
                ['precision', 'recall', 'f1 score'],
                labels=labels_np[higher_f1s].tolist(),
                title='{} Multiclass Precision / Recall / '
                      'F1 Score best {} classes {}'.format(
                    model_name_name, k, field)
            )
            lower_f1s = sorted_indices[:k]
            visualization_utils.compare_classifiers_multiclass_multimetric_plot(
                [p_np[lower_f1s],
                 r_np[lower_f1s],
                 f1_np[lower_f1s]],
                ['precision', 'recall', 'f1 score'],
                labels=labels_np[lower_f1s].tolist(),
                title='{} Multiclass Precision / Recall / F1 Score worst '
                      'k classes {}'.format(model_name_name, k, field)
            )

            visualization_utils.compare_classifiers_multiclass_multimetric_plot(
                [p_np[sorted_indices[::-1]],
                 r_np[sorted_indices[::-1]],
                 f1_np[sorted_indices[::-1]]],
                ['precision', 'recall', 'f1 score'],
                labels=labels_np[sorted_indices[::-1]].tolist(),
                title='{} Multiclass Precision / Recall / F1 Score '
                      '{} sorted'.format(model_name_name, field)
            )

            logging.info('\n')
            logging.info(model_name_name)
            tmp_str = '{0} best 5 classes: '.format(field)
            tmp_str += '{}'
            logging.info(tmp_str.format(higher_f1s))
            logging.info(f1_np[higher_f1s])
            tmp_str = '{0} worst 5 classes: '.format(field)
            tmp_str += '{}'
            logging.info(tmp_str.format(lower_f1s))
            logging.info(f1_np[lower_f1s])
            tmp_str = '{0} number of classes with f1 score > 0: '.format(field)
            tmp_str += '{}'
            logging.info(tmp_str.format(np.sum(f1_np > 0)))
            tmp_str = '{0} number of classes with f1 score = 0: '.format(field)
            tmp_str += '{}'
            logging.info(tmp_str.format(np.sum(f1_np == 0)))


def compare_classifiers_predictions(
        predictions,
        ground_truth,
        field,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(predictions) < 2:
        logging.error('No predictions provided')
        return

    ground_truth_fn = ground_truth
    ground_truth_field = field
    predictions_1_fn = predictions[0]
    predictions_2_fn = predictions[1]
    name_c1 = (
        model_names[0] if model_names is not None and len(model_names) > 0
        else 'c1')
    name_c2 = (
        model_names[1] if model_names is not None and len(model_names) > 1
        else 'c2')

    gt = load_from_file(ground_truth_fn, ground_truth_field)
    pred_c1 = load_from_file(predictions_1_fn, dtype=int)
    pred_c2 = load_from_file(predictions_2_fn, dtype=int)

    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit
        pred_c1[pred_c1 > labels_limit] = labels_limit
        pred_c2[pred_c2 > labels_limit] = labels_limit

    # DOTO all shadows built in name - come up with a more descriptive name
    all = len(gt)
    if all == 0:
        logging.error('No labels in the ground truth')
        return

    both_right = 0
    both_wrong_same = 0
    both_wrong_different = 0
    c1_right_c2_wrong = 0
    c1_wrong_c2_right = 0

    for i in range(all):
        if gt[i] == pred_c1[i] and gt[i] == pred_c2[i]:
            both_right += 1
        elif gt[i] != pred_c1[i] and gt[i] != pred_c2[i]:
            if pred_c1[i] == pred_c2[i]:
                both_wrong_same += 1
            else:
                both_wrong_different += 1
        elif gt[i] == pred_c1[i] and gt[i] != pred_c2[i]:
            c1_right_c2_wrong += 1
        elif gt[i] != pred_c1[i] and gt[i] == pred_c2[i]:
            c1_wrong_c2_right += 1

    one_right = c1_right_c2_wrong + c1_wrong_c2_right
    both_wrong = both_wrong_same + both_wrong_different

    logging.info('Test datapoints: {}'.format(all))
    logging.info(
        'Both right: {} {:.2f}%'.format(both_right, 100 * both_right / all))
    logging.info(
        'One right: {} {:.2f}%'.format(one_right, 100 * one_right / all))
    logging.info(
        '  {} right / {} wrong: {} {:.2f}% {:.2f}%'.format(
            name_c1,
            name_c2,
            c1_right_c2_wrong,
            100 * c1_right_c2_wrong / all,
            100 * c1_right_c2_wrong / one_right if one_right > 0 else 0
        )
    )
    logging.info(
        '  {} wrong / {} right: {} {:.2f}% {:.2f}%'.format(
            name_c1,
            name_c2,
            c1_wrong_c2_right,
            100 * c1_wrong_c2_right / all,
            100 * c1_wrong_c2_right / one_right if one_right > 0 else 0
        )
    )
    logging.info(
        'Both wrong: {} {:.2f}%'.format(both_wrong, 100 * both_wrong / all)
    )
    logging.info('  same prediction: {} {:.2f}% {:.2f}%'.format(
        both_wrong_same,
        100 * both_wrong_same / all,
        100 * both_wrong_same / both_wrong if both_wrong > 0 else 0
    )
    )
    logging.info('  different prediction: {} {:.2f}% {:.2f}%'.format(
        both_wrong_different,
        100 * both_wrong_different / all,
        100 * both_wrong_different / both_wrong if both_wrong > 0 else 0
    )
    )

    visualization_utils.donut(
        [both_right, one_right, both_wrong],
        ['both right', 'one right', 'both wrong'],
        [both_right, c1_right_c2_wrong, c1_wrong_c2_right, both_wrong_same,
         both_wrong_different],
        ['both right',
         '{} right / {} wrong'.format(name_c1, name_c2),
         '{} wrong / {} right'.format(name_c1, name_c2),
         'same prediction', 'different prediction'],
        [0, 1, 1, 2, 2],
        title='{} vs {}'.format(name_c1, name_c2)
    )


def compare_classifiers_predictions_distribution(
        predictions,
        ground_truth,
        field,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(predictions) < 1:
        logging.error('No predictions provided')
        return

    ground_truth = load_from_file(ground_truth, field)
    if labels_limit > 0:
        ground_truth[ground_truth > labels_limit] = labels_limit

    predictions = [load_from_file(predictions_fn, dtype=int)
                   for predictions_fn in predictions]
    if labels_limit > 0:
        for i in range(len(predictions)):
            predictions[i][predictions[i] > labels_limit] = labels_limit

    counts_gt = np.bincount(ground_truth)
    prob_gt = counts_gt / counts_gt.sum()

    counts_predictions = [np.bincount(alg_predictions)
                          for alg_predictions in predictions]

    prob_predictions = [alg_count_prediction / alg_count_prediction.sum()
                        for alg_count_prediction in counts_predictions]

    visualization_utils.radar_chart(prob_gt, prob_predictions, model_names)


def confidence_thresholding(
        probabilities,
        ground_truth,
        field,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    thresholds = [t / 100 for t in range(0, 101, 5)]

    accuracies = []
    dataset_kept = []

    for i, prob in enumerate(probs):

        if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
            prob_limit = prob[:, :labels_limit + 1]
            prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
            prob = prob_limit

        max_prob = np.max(prob, axis=1)
        predictions = np.argmax(prob, axis=1)

        accuracies_alg = []
        dataset_kept_alg = []

        for threshold in thresholds:
            threshold = threshold if threshold < 1 else 0.999
            filtered_indices = max_prob >= threshold
            filtered_gt = gt[filtered_indices]
            filtered_predictions = predictions[filtered_indices]
            accuracy = (
                    (filtered_gt == filtered_predictions).sum() /
                    len(filtered_gt)
            )

            accuracies_alg.append(accuracy)
            dataset_kept_alg.append(len(filtered_gt) / len(gt))

        accuracies.append(accuracies_alg)
        dataset_kept.append(dataset_kept_alg)

    visualization_utils.confidence_fitlering_plot(
        thresholds,
        accuracies,
        dataset_kept,
        model_names,
        title='Confidence_Thresholding'
    )


def confidence_thresholding_data_vs_acc(
        probabilities,
        ground_truth,
        field,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    thresholds = [t / 100 for t in range(0, 101, 5)]

    accuracies = []
    dataset_kept = []

    for i, prob in enumerate(probs):

        if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
            prob_limit = prob[:, :labels_limit + 1]
            prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
            prob = prob_limit

        max_prob = np.max(prob, axis=1)
        predictions = np.argmax(prob, axis=1)

        accuracies_alg = []
        dataset_kept_alg = []

        for threshold in thresholds:
            threshold = threshold if threshold < 1 else 0.999
            filtered_indices = max_prob >= threshold
            filtered_gt = gt[filtered_indices]
            filtered_predictions = predictions[filtered_indices]
            accuracy = ((filtered_gt == filtered_predictions).sum() /
                        len(filtered_gt))

            accuracies_alg.append(accuracy)
            dataset_kept_alg.append(len(filtered_gt) / len(gt))

        accuracies.append(accuracies_alg)
        dataset_kept.append(dataset_kept_alg)

    visualization_utils.confidence_fitlering_data_vs_acc_plot(
        accuracies,
        dataset_kept,
        model_names,
        title='Confidence_Thresholding (Data vs Accuracy)'
    )


def confidence_thresholding_data_vs_acc_subset(
        probabilities,
        ground_truth,
        field,
        top_n_classes,
        labels_limit,
        subset,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    k = top_n_classes[0]
    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    thresholds = [t / 100 for t in range(0, 101, 5)]

    accuracies = []
    dataset_kept = []

    subset_indices = gt > 0
    gt_subset = gt
    if subset == 'ground_truth':
        subset_indices = gt < k
        gt_subset = gt[subset_indices]
        logging.info('Subset is {:.2f}% of the data'.format(
            len(gt_subset) / len(gt) * 100)
        )

    for i, prob in enumerate(probs):

        if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
            prob_limit = prob[:, :labels_limit + 1]
            prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
            prob = prob_limit

        if subset == PREDICTIONS:
            subset_indices = np.argmax(prob, axis=1) < k
            gt_subset = gt[subset_indices]
            logging.info(
                'Subset for model_name {} is {:.2f}% of the data'.format(
                    model_names[i] if model_names and i < len(
                        model_names) else i,
                    len(gt_subset) / len(gt) * 100
                )
            )

        prob_subset = prob[subset_indices]

        max_prob = np.max(prob_subset, axis=1)
        predictions = np.argmax(prob_subset, axis=1)

        accuracies_alg = []
        dataset_kept_alg = []

        for threshold in thresholds:
            threshold = threshold if threshold < 1 else 0.999
            filtered_indices = max_prob >= threshold
            filtered_gt = gt_subset[filtered_indices]
            filtered_predictions = predictions[filtered_indices]
            accuracy = ((filtered_gt == filtered_predictions).sum() /
                        len(filtered_gt))

            accuracies_alg.append(accuracy)
            dataset_kept_alg.append(len(filtered_gt) / len(gt))

        accuracies.append(accuracies_alg)
        dataset_kept.append(dataset_kept_alg)

    visualization_utils.confidence_fitlering_data_vs_acc_plot(
        accuracies,
        dataset_kept,
        model_names,
        title='Confidence_Thresholding (Data vs Accuracy)'
    )


def confidence_thresholding_data_vs_acc_subset_per_class(
        probabilities,
        ground_truth,
        ground_truth_metadata,
        field,
        top_n_classes,
        labels_limit,
        subset,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    metadata = load_json(ground_truth_metadata)
    k = top_n_classes[0]
    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    thresholds = [t / 100 for t in range(0, 101, 5)]

    for curr_k in range(k):
        accuracies = []
        dataset_kept = []

        subset_indices = gt > 0
        gt_subset = gt
        if subset == 'ground_truth':
            subset_indices = gt == curr_k
            gt_subset = gt[subset_indices]
            logging.info('Subset is {:.2f}% of the data'.format(
                len(gt_subset) / len(gt) * 100)
            )

        for i, prob in enumerate(probs):

            if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
                prob_limit = prob[:, :labels_limit + 1]
                prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
                prob = prob_limit

            if subset == PREDICTIONS:
                subset_indices = np.argmax(prob, axis=1) == curr_k
                gt_subset = gt[subset_indices]
                logging.info(
                    'Subset for model_name {} is {:.2f}% of the data'.format(
                        model_names[i] if model_names and i < len(
                            model_names) else i,
                        len(gt_subset) / len(gt) * 100
                    )
                )

            prob_subset = prob[subset_indices]

            max_prob = np.max(prob_subset, axis=1)
            predictions = np.argmax(prob_subset, axis=1)

            accuracies_alg = []
            dataset_kept_alg = []

            for threshold in thresholds:
                threshold = threshold if threshold < 1 else 0.999
                filtered_indices = max_prob >= threshold
                filtered_gt = gt_subset[filtered_indices]
                filtered_predictions = predictions[filtered_indices]
                accuracy = ((filtered_gt == filtered_predictions).sum() /
                            len(filtered_gt) if len(filtered_gt) > 0 else 0)

                accuracies_alg.append(accuracy)
                dataset_kept_alg.append(len(filtered_gt) / len(gt))

            accuracies.append(accuracies_alg)
            dataset_kept.append(dataset_kept_alg)

        visualization_utils.confidence_fitlering_data_vs_acc_plot(
            accuracies, dataset_kept, model_names,
            decimal_digits=2,
            title='Confidence_Thresholding (Data vs Accuracy) '
                  'for class {}'.format(metadata[field]['idx2str'][curr_k])
        )


def confidence_thresholding_2thresholds_2d(
        probabilities,
        ground_truth,
        threshold_fields,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 2:
        logging.error('Not enough probabilities provided, two are needed')
        return
    if len(probabilities) > 2:
        logging.error('Too many probabilities provided, only two are needed')
        return
    if len(threshold_fields) < 2:
        logging.error('Not enough threshold fields provided, two are needed')
        return
    if len(threshold_fields) > 2:
        logging.error('Too many threshold fields provided, only two are needed')
        return

    gt_1 = load_from_file(ground_truth, threshold_fields[0])
    gt_2 = load_from_file(ground_truth, threshold_fields[1])

    if labels_limit > 0:
        gt_1[gt_1 > labels_limit] = labels_limit
        gt_2[gt_2 > labels_limit] = labels_limit

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    thresholds = [t / 100 for t in range(0, 101, 5)]
    fixed_step_coverage = thresholds
    name_t1 = '{} threshold'.format(threshold_fields[0])
    name_t2 = '{} threshold'.format(threshold_fields[1])

    accuracies = []
    dataset_kept = []
    interps = []
    table = [[name_t1, name_t2, 'coverage', ACCURACY]]

    if labels_limit > 0 and probs[0].shape[1] > labels_limit + 1:
        prob_limit = probs[0][:, :labels_limit + 1]
        prob_limit[:, labels_limit] = probs[0][:, labels_limit:].sum(1)
        probs[0] = prob_limit

    if labels_limit > 0 and probs[1].shape[1] > labels_limit + 1:
        prob_limit = probs[1][:, :labels_limit + 1]
        prob_limit[:, labels_limit] = probs[1][:, labels_limit:].sum(1)
        probs[1] = prob_limit

    max_prob_1 = np.max(probs[0], axis=1)
    predictions_1 = np.argmax(probs[0], axis=1)

    max_prob_2 = np.max(probs[1], axis=1)
    predictions_2 = np.argmax(probs[1], axis=1)

    for threshold_1 in thresholds:
        threshold_1 = threshold_1 if threshold_1 < 1 else 0.999
        curr_accuracies = []
        curr_dataset_kept = []

        for threshold_2 in thresholds:
            threshold_2 = threshold_2 if threshold_2 < 1 else 0.999

            filtered_indices = np.logical_and(
                max_prob_1 >= threshold_1,
                max_prob_2 >= threshold_2
            )

            filtered_gt_1 = gt_1[filtered_indices]
            filtered_predictions_1 = predictions_1[filtered_indices]
            filtered_gt_2 = gt_2[filtered_indices]
            filtered_predictions_2 = predictions_2[filtered_indices]

            coverage = len(filtered_gt_1) / len(gt_1)
            accuracy = (
                           np.logical_and(
                               filtered_gt_1 == filtered_predictions_1,
                               filtered_gt_2 == filtered_predictions_2
                           )
                       ).sum() / len(filtered_gt_1)

            curr_accuracies.append(accuracy)
            curr_dataset_kept.append(coverage)
            table.append([threshold_1, threshold_2, coverage, accuracy])

        accuracies.append(curr_accuracies)
        dataset_kept.append(curr_dataset_kept)
        interps.append(
            np.interp(
                fixed_step_coverage,
                list(reversed(curr_dataset_kept)),
                list(reversed(curr_accuracies)),
                left=1,
                right=0
            )
        )

    logging.info('CSV table')
    for row in table:
        logging.info(','.join([str(e) for e in row]))

    # ===========#
    # Multiline #
    # ===========#
    visualization_utils.confidence_fitlering_data_vs_acc_multiline_plot(
        accuracies,
        dataset_kept,
        model_names,
        title='Coverage vs Accuracy, two thresholds'
    )

    # ==========#
    # Max line #
    # ==========#
    max_accuracies = np.amax(np.array(interps), 0)
    visualization_utils.confidence_fitlering_data_vs_acc_plot(
        [max_accuracies],
        [thresholds],
        model_names,
        title='Coverage vs Accuracy, two thresholds'
    )

    # ==========================#
    # Max line with thresholds #
    # ==========================#
    acc_matrix = np.array(accuracies)
    cov_matrix = np.array(dataset_kept)
    t1_maxes = [1]
    t2_maxes = [1]
    for i in range(len(fixed_step_coverage) - 1):
        lower = fixed_step_coverage[i]
        upper = fixed_step_coverage[i + 1]
        indices = np.logical_and(cov_matrix >= lower, cov_matrix < upper)
        selected_acc = acc_matrix.copy()
        selected_acc[np.logical_not(indices)] = -1
        threshold_indices = np.unravel_index(np.argmax(selected_acc, axis=None),
                                             selected_acc.shape)
        t1_maxes.append(thresholds[threshold_indices[0]])
        t2_maxes.append(thresholds[threshold_indices[1]])
    model_name = model_names[0] if model_names is not None and len(
        model_names) > 0 else ''
    visualization_utils.confidence_fitlering_data_vs_acc_plot(
        [max_accuracies, t1_maxes, t2_maxes],
        [fixed_step_coverage, fixed_step_coverage, fixed_step_coverage],
        model_names=[model_name + ' accuracy', name_t1, name_t2],
        dotted=[False, True, True],
        y_label='',
        title='Coverage vs Accuracy & Threshold'
    )


def confidence_thresholding_2thresholds_3d(
        probabilities,
        ground_truth,
        threshold_fields,
        labels_limit,
        **kwargs
):
    if len(probabilities) < 2:
        logging.error('Not enough probabilities provided, two are needed')
        return
    if len(probabilities) > 2:
        logging.error('Too many probabilities provided, only two are needed')
        return
    if len(threshold_fields) < 2:
        logging.error('Not enough threshold fields provided, two are needed')
        return
    if len(threshold_fields) > 2:
        logging.error('Too many threshold fields provided, only two are needed')
        return

    gt_1 = load_from_file(ground_truth, threshold_fields[0])
    gt_2 = load_from_file(ground_truth, threshold_fields[1])

    if labels_limit > 0:
        gt_1[gt_1 > labels_limit] = labels_limit
        gt_2[gt_2 > labels_limit] = labels_limit

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    thresholds = [t / 100 for t in range(0, 101, 5)]

    accuracies = []
    dataset_kept = []

    if labels_limit > 0 and probs[0].shape[1] > labels_limit + 1:
        prob_limit = probs[0][:, :labels_limit + 1]
        prob_limit[:, labels_limit] = probs[0][:, labels_limit:].sum(1)
        probs[0] = prob_limit

    if labels_limit > 0 and probs[1].shape[1] > labels_limit + 1:
        prob_limit = probs[1][:, :labels_limit + 1]
        prob_limit[:, labels_limit] = probs[1][:, labels_limit:].sum(1)
        probs[1] = prob_limit

    max_prob_1 = np.max(probs[0], axis=1)
    predictions_1 = np.argmax(probs[0], axis=1)

    max_prob_2 = np.max(probs[1], axis=1)
    predictions_2 = np.argmax(probs[1], axis=1)

    for threshold_1 in thresholds:
        threshold_1 = threshold_1 if threshold_1 < 1 else 0.999
        curr_accuracies = []
        curr_dataset_kept = []

        for threshold_2 in thresholds:
            threshold_2 = threshold_2 if threshold_2 < 1 else 0.999

            filtered_indices = np.logical_and(
                max_prob_1 >= threshold_1,
                max_prob_2 >= threshold_2
            )

            filtered_gt_1 = gt_1[filtered_indices]
            filtered_predictions_1 = predictions_1[filtered_indices]
            filtered_gt_2 = gt_2[filtered_indices]
            filtered_predictions_2 = predictions_2[filtered_indices]

            accuracy = (
                           np.logical_and(
                               filtered_gt_1 == filtered_predictions_1,
                               filtered_gt_2 == filtered_predictions_2
                           )
                       ).sum() / len(filtered_gt_1)

            curr_accuracies.append(accuracy)
            curr_dataset_kept.append(len(filtered_gt_1) / len(gt_1))

        accuracies.append(curr_accuracies)
        dataset_kept.append(curr_dataset_kept)

    visualization_utils.confidence_fitlering_3d_plot(
        np.array(thresholds),
        np.array(thresholds),
        np.array(accuracies),
        np.array(dataset_kept),
        threshold_fields,
        title='Confidence_Thresholding, two thresholds'
    )


def binary_threshold_vs_metric(
        probabilities,
        ground_truth,
        field,
        metrics,
        positive_label=1,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    gt = load_from_file(ground_truth, field)

    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    thresholds = [t / 100 for t in range(0, 101, 5)]

    supported_metrics = {'f1', 'precision', 'recall', 'accuracy'}

    for metric in metrics:

        if metric not in supported_metrics:
            logging.error("Metric {} not supported".format(metric))
            continue

        scores = []

        for i, prob in enumerate(probs):

            scores_alg = []

            if len(prob.shape) == 2:
                if prob.shape[1] > positive_label:
                    prob = prob[:, positive_label]
                else:
                    raise Exception(
                        'the specified positive label {} is not '
                        'present in the probabilities'.format(
                            positive_label
                        )
                    )

            for threshold in thresholds:
                threshold = threshold if threshold < 1 else 0.99

                t_gt = gt[prob >= threshold]
                predictions = prob >= threshold
                t_predictions = predictions[prob >= threshold]

                if metric == 'f1':
                    metric_score = sklearn.metrics.f1_score(
                        t_gt,
                        t_predictions
                    )
                elif metric == 'precision':
                    metric_score = sklearn.metrics.precision_score(
                        t_gt,
                        t_predictions
                    )
                elif metric == 'recall':
                    metric_score = sklearn.metrics.recall_score(
                        t_gt,
                        t_predictions
                    )
                elif metric == ACCURACY:
                    metric_score = sklearn.metrics.accuracy_score(
                        t_gt,
                        t_predictions
                    )

                scores_alg.append(metric_score)

            scores.append(scores_alg)

        visualization_utils.threshold_vs_metric_plot(
            thresholds,
            scores,
            model_names,
            title='Binary threshold vs {}'.format(metric)
        )


def roc_curves(
        probabilities,
        ground_truth,
        field,
        positive_label=1,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    gt = load_from_file(ground_truth, field)
    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    fpr_tprs = []

    for i, prob in enumerate(probs):
        if len(prob.shape) > 1:
            prob = prob[:, positive_label]
        fpr, tpr, _ = sklearn.metrics.roc_curve(
            gt, prob,
            pos_label=positive_label
        )
        fpr_tprs.append((fpr, tpr))

    visualization_utils.roc_curves(fpr_tprs, model_names, title='ROC curves')


def roc_curves_from_prediction_statistics(prediction_statistics, field,
                                          model_names=None, **kwargs):
    if len(prediction_statistics) < 1:
        logging.error('No prediction_statistics provided')
        return

    prediction_statistics_per_model_name = [load_json(prediction_statistics_f)
                                            for prediction_statistics_f in
                                            prediction_statistics]
    fpr_tprs = []
    for curr_prediction_statistics in prediction_statistics_per_model_name:
        fpr = curr_prediction_statistics[field]['roc_curve'][
            'false_positive_rate']
        tpr = curr_prediction_statistics[field]['roc_curve'][
            'true_positive_rate']
        fpr_tprs.append((fpr, tpr))

    visualization_utils.roc_curves(fpr_tprs, model_names, title='ROC curves')


def calibration_1_vs_all(
        probabilities,
        ground_truth,
        field,
        top_n_classes,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit
    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]
    for i, prob in enumerate(probs):
        if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
            prob_limit = prob[:, :labels_limit + 1]
            prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
            probs[i] = prob_limit

    num_classes = max(gt)

    brier_scores = []

    classes = (min(num_classes, top_n_classes[0]) if top_n_classes[0] > 0
               else num_classes)

    for class_idx in range(classes):
        fraction_positives_class = []
        mean_predicted_vals_class = []
        probs_class = []
        brier_scores_class = []
        for prob in probs:
            # ground_truth is an vector of integers, each integer is a class
            # index to have a [0,1] vector we have to check if the value equals
            # the input class index and convert the resulting boolean vector
            # into an integer vector probabilities is a n x c matrix, n is the
            # number of datapoints and c number of classes; its values are the
            # probabilities of the ith datapoint to be classified as belonging
            # to the jth class according to the learned model. For this reason
            # we need to take only the column of predictions that is about the
            # class we are interested in, the input class index

            gt_class = (gt == class_idx).astype(int)
            prob_class = prob[:, class_idx]

            (
                curr_fraction_positives,
                curr_mean_predicted_vals
            ) = calibration_curve(gt_class, prob_class, n_bins=21)

            fraction_positives_class.append(curr_fraction_positives)
            mean_predicted_vals_class.append(curr_mean_predicted_vals)
            probs_class.append(prob[:, class_idx])
            brier_scores_class.append(
                brier_score_loss(
                    gt_class,
                    prob_class, pos_label=1
                )
            )

        brier_scores.append(brier_scores_class)
        visualization_utils.calibration_plot(
            fraction_positives_class,
            mean_predicted_vals_class,
            model_names
        )
        visualization_utils.predictions_distribution_plot(
            probs_class,
            model_names
        )

    visualization_utils.brier_plot(np.array(brier_scores), model_names)


def calibration_multiclass(
        probabilities,
        ground_truth,
        field,
        labels_limit,
        model_names=None,
        **kwargs
):
    if len(probabilities) < 1:
        logging.error('No probabilities provided')
        return

    gt = load_from_file(ground_truth, field)
    if labels_limit > 0:
        gt[gt > labels_limit] = labels_limit

    prob_classes = 0
    probs = [load_from_file(probs_fn, dtype=float)
             for probs_fn in probabilities]

    for i, prob in enumerate(probs):
        if labels_limit > 0 and prob.shape[1] > labels_limit + 1:
            prob_limit = prob[:, :labels_limit + 1]
            prob_limit[:, labels_limit] = prob[:, labels_limit:].sum(1)
            probs[i] = prob_limit
        if probs[i].shape[1] > prob_classes:
            prob_classes = probs[i].shape[1]

    gt_one_hot_dim_2 = max(prob_classes, max(gt) + 1)
    gt_one_hot = np.zeros((len(gt), gt_one_hot_dim_2))
    gt_one_hot[np.arange(len(gt)), gt] = 1
    gt_one_hot_flat = gt_one_hot.flatten()

    fraction_positives = []
    mean_predicted_vals = []
    brier_scores = []
    for prob in probs:
        # flatten probabilities to be compared to flatten ground truth
        prob_flat = prob.flatten()
        curr_fraction_positives, curr_mean_predicted_vals = calibration_curve(
            gt_one_hot_flat,
            prob_flat,
            n_bins=21
        )
        fraction_positives.append(curr_fraction_positives)
        mean_predicted_vals.append(curr_mean_predicted_vals)
        brier_scores.append(
            brier_score_loss(
                gt_one_hot_flat,
                prob_flat,
                pos_label=1
            )
        )

    visualization_utils.calibration_plot(
        fraction_positives,
        mean_predicted_vals,
        model_names
    )
    visualization_utils.compare_classifiers_plot(
        [brier_scores],
        ['brier'],
        model_names,
        adaptive=True,
        decimals=8
    )
    for i, brier_score in enumerate(brier_scores):
        if i < len(model_names):
            format_str = '{}: '.format(model_names[i])
            format_str += '{}'
        else:
            format_str = '{}'
        logging.info(format_str.format(brier_score))


def confusion_matrix(
        prediction_statistics,
        ground_truth_metadata,
        field,
        top_n_classes,
        normalize,
        model_names=None,
        **kwargs
):
    if len(prediction_statistics) < 1:
        logging.error('No prediction_statistics provided')
        return

    metadata = load_json(ground_truth_metadata)
    prediction_statistics_per_model_name = [load_json(prediction_statistics_f)
                                            for prediction_statistics_f in
                                            prediction_statistics]

    fields_set = set()
    for ls in prediction_statistics_per_model_name:
        for key in ls:
            fields_set.add(key)
    fields = [field] if field is not None and len(field) > 0 else fields_set

    for i, prediction_statistics in enumerate(
            prediction_statistics_per_model_name):
        for field in fields:
            if 'confusion_matrix' in prediction_statistics[field]:
                confusion_matrix = np.array(
                    prediction_statistics[field]['confusion_matrix']
                )
                model_name_name = model_names[i] if (
                        model_names is not None and i < len(model_names)
                ) else ''

                if field in metadata and 'idx2str' in metadata[field]:
                    labels = metadata[field]['idx2str']
                else:
                    labels = list(range(len(confusion_matrix)))

                for k in top_n_classes:
                    k = (min(k, confusion_matrix.shape[0])
                         if k > 0 else confusion_matrix.shape[0])
                    cm = confusion_matrix[:k, :k]
                    if normalize:
                        with np.errstate(divide='ignore', invalid='ignore'):
                            cm_norm = np.true_divide(cm,
                                                     cm.sum(1)[:, np.newaxis])
                            cm_norm[cm_norm == np.inf] = 0
                            cm_norm = np.nan_to_num(cm_norm)
                        cm = cm_norm
                    visualization_utils.confusion_matrix_plot(
                        cm,
                        labels[:k],
                        field=field
                    )

                    entropies = []
                    for row in cm:
                        if np.count_nonzero(row) > 0:
                            entropies.append(entropy(row))
                        else:
                            entropies.append(0)
                    class_entropy = np.array(entropies)
                    class_desc_entropy = np.argsort(class_entropy)[::-1]
                    desc_entropy = class_entropy[class_desc_entropy]
                    visualization_utils.bar_plot(
                        class_desc_entropy,
                        desc_entropy,
                        labels=[labels[i] for i in class_desc_entropy],
                        title='Classes ranked by entropy of '
                              'Confusion Matrix row'
                    )


def frequency_vs_f1(
        prediction_statistics,
        ground_truth_metadata,
        field,
        top_n_classes,
        model_names=None,
        **kwargs
):
    if len(prediction_statistics) < 1:
        logging.error('No prediction_statistics provided')
        return

    metadata = load_json(ground_truth_metadata)
    prediction_statistics_per_model_name = [load_json(prediction_statistics_f)
                                            for prediction_statistics_f in
                                            prediction_statistics]
    k = top_n_classes[0]

    fields_set = set()
    for ls in prediction_statistics_per_model_name:
        for key in ls:
            fields_set.add(key)
    fields = [field] if field is not None and len(field) > 0 else fields_set

    for i, prediction_statistics in enumerate(
            prediction_statistics_per_model_name):
        for field in fields:
            model_name_name = (model_names[i]
                               if model_names is not None and i < len(
                model_names)
                               else '')
            per_class_stats = prediction_statistics[field]['per_class_stats']
            f1_scores = []
            labels = []
            class_names = metadata[field]['idx2str']
            if k > 0:
                class_names = class_names[:k]
            for class_name in class_names:
                class_stats = per_class_stats[class_name]
                f1_scores.append(class_stats['f1_score'])
                labels.append(class_name)

            f1_np = np.nan_to_num(np.array(f1_scores, dtype=np.float32))
            f1_sorted_indices = f1_np.argsort()

            field_frequency_dict = {
                metadata[field]['str2idx'][key]: val
                for key, val in metadata[field]['str2freq'].items()
            }
            field_frequency_np = np.array(
                [field_frequency_dict[class_id]
                 for class_id in sorted(field_frequency_dict)],
                dtype=np.int32
            )

            field_frequency_reordered = field_frequency_np[
                                            f1_sorted_indices[::-1]
                                        ][:len(f1_sorted_indices)]
            f1_reordered = f1_np[f1_sorted_indices[::-1]][
                           :len(f1_sorted_indices)]

            visualization_utils.double_axis_line_plot(
                f1_reordered,
                field_frequency_reordered,
                'F1 score',
                'frequency',
                labels=labels,
                title='{} F1 Score vs Frequency {}'.format(
                    model_name_name,
                    field
                )
            )

            frequency_sorted_indices = field_frequency_np.argsort()
            field_frequency_reordered = field_frequency_np[
                                            frequency_sorted_indices[::-1]
                                        ][:len(f1_sorted_indices)]

            f1_reordered = np.zeros(len(field_frequency_reordered))

            for idx in frequency_sorted_indices[::-1]:
                if idx < len(f1_np):
                    f1_reordered[idx] = f1_np[idx]

            visualization_utils.double_axis_line_plot(
                field_frequency_reordered,
                f1_reordered,
                'frequency',
                'F1 score',
                labels=labels,
                title='{} F1 Score vs Frequency {}'.format(
                    model_name_name,
                    field
                )
            )


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description='This script analyzes results and shows some nice plots.',
        prog='ludwig visualize',
        usage='%(prog)s [options]')

    parser.add_argument('-d', '--data_csv', help='raw data file')
    parser.add_argument('-g', '--ground_truth', help='ground truth file')
    parser.add_argument(
        '-gm',
        '--ground_truth_metadata',
        help='input metadata JSON file'
    )

    parser.add_argument(
        '-v',
        '--visualization',
        default='confidence_thresholding',
        choices=['learning_curves',
                 'compare_performance',
                 'compare_classifiers_performance_from_prob',
                 'compare_classifiers_performance_from_pred',
                 'compare_classifiers_performance_subset',
                 'compare_classifiers_performance_changing_k',
                 'compare_classifiers_multiclass_multimetric',
                 'compare_classifiers_predictions',
                 'compare_classifiers_predictions_distribution',
                 'confidence_thresholding',
                 'confidence_thresholding_data_vs_acc',
                 'confidence_thresholding_data_vs_acc_subset',
                 'confidence_thresholding_data_vs_acc_subset_per_class',
                 'confidence_thresholding_2thresholds_2d',
                 'confidence_thresholding_2thresholds_3d',
                 'binary_threshold_vs_metric',
                 'roc_curves',
                 'roc_curves_from_prediction_statistics',
                 'calibration_1_vs_all',
                 'calibration_multiclass',
                 'confusion_matrix',
                 'frequency_vs_f1'],
        help='type of visualization'
    )

    parser.add_argument(
        '-f',
        '--field',
        default=[],
        help='field containing ground truth'
    )
    parser.add_argument(
        '-tf',
        '--threshold_fields',
        default=[],
        nargs='+',
        help='fields for 2d threshold'
    )
    parser.add_argument(
        '-pred',
        '--predictions',
        default=[],
        nargs='+',
        type=str,
        help='predictions files'
    )
    parser.add_argument(
        '-prob',
        '--probabilities',
        default=[],
        nargs='+',
        type=str,
        help='probabilities files'
    )
    parser.add_argument(
        '-ts',
        '--training_statistics',
        default=[],
        nargs='+',
        type=str,
        help='training stats files'
    )
    parser.add_argument(
        '-ps',
        '--prediction_statistics',
        default=[],
        nargs='+',
        type=str,
        help='test stats files'
    )
    parser.add_argument(
        '-mn',
        '--model_names',
        default=[],
        nargs='+',
        type=str,
        help='names of the models to use as labels'
    )
    parser.add_argument(
        '-tn',
        '--top_n_classes',
        default=[0],
        nargs='+',
        type=int,
        help='number of classes to plot'
    )
    parser.add_argument(
        '-k',
        '--top_k',
        default=3,
        type=int,
        help='number of elements in the ranklist to consider'
    )
    parser.add_argument(
        '-ll',
        '--labels_limit',
        default=0,
        type=int,
        help='maximum numbers of labels. '
             'If labels in dataset are higher than this number, "rare" label'
    )
    parser.add_argument(
        '-ss',
        '--subset',
        default='ground_truth',
        choices=['ground_truth', PREDICTIONS],
        help='type of subset filtering'
    )
    parser.add_argument(
        '-n',
        '--normalize',
        action='store_true',
        default=False,
        help='normalize rows in confusion matrix'
    )
    parser.add_argument(
        '-m',
        '--metrics',
        default=['f1'],
        nargs='+',
        type=str,
        help='metrics to dispay in threshold_vs_metric'
    )
    parser.add_argument(
        '-pl',
        '--positive_label',
        type=int,
        default=1,
        help='label of the positive class for the roc curve'
    )
    parser.add_argument(
        '-l',
        '--logging_level',
        default='info',
        help='the level of logging to use',
        choices=['critical', 'error', 'warning', 'info', 'debug', 'notset']
    )

    args = parser.parse_args(sys_argv)

    logging.basicConfig(
        stream=sys.stdout,
        level=logging_level_registry[args.logging_level],
        format='%(message)s'
    )

    if args.visualization == 'compare_performance':
        compare_performance(**vars(args))
    elif args.visualization == 'compare_classifiers_performance_from_prob':
        compare_classifiers_performance_from_prob(**vars(args))
    elif args.visualization == 'compare_classifiers_performance_from_pred':
        compare_classifiers_performance_from_pred(**vars(args))
    elif args.visualization == 'compare_classifiers_performance_subset':
        compare_classifiers_performance_subset(**vars(args))
    elif args.visualization == 'compare_classifiers_performance_changing_k':
        compare_classifiers_performance_changing_k(**vars(args))
    elif args.visualization == 'compare_classifiers_multiclass_multimetric':
        compare_classifiers_multiclass_multimetric(**vars(args))
    elif args.visualization == 'compare_classifiers_predictions':
        compare_classifiers_predictions(**vars(args))
    elif args.visualization == 'compare_classifiers_predictions_distribution':
        compare_classifiers_predictions_distribution(**vars(args))
    elif args.visualization == 'confidence_thresholding':
        confidence_thresholding(**vars(args))
    elif args.visualization == 'confidence_thresholding_data_vs_acc':
        confidence_thresholding_data_vs_acc(**vars(args))
    elif args.visualization == 'confidence_thresholding_data_vs_acc_subset':
        confidence_thresholding_data_vs_acc_subset(**vars(args))
    elif (args.visualization ==
          'confidence_thresholding_data_vs_acc_subset_per_class'):
        confidence_thresholding_data_vs_acc_subset_per_class(**vars(args))
    elif args.visualization == 'confidence_thresholding_2thresholds_2d':
        confidence_thresholding_2thresholds_2d(**vars(args))
    elif args.visualization == 'confidence_thresholding_2thresholds_3d':
        confidence_thresholding_2thresholds_3d(**vars(args))
    elif args.visualization == 'binary_threshold_vs_metric':
        binary_threshold_vs_metric(**vars(args))
    elif args.visualization == 'roc_curves':
        roc_curves(**vars(args))
    elif args.visualization == 'roc_curves_from_prediction_statistics':
        roc_curves_from_prediction_statistics(**vars(args))
    elif args.visualization == 'calibration_1_vs_all':
        calibration_1_vs_all(**vars(args))
    elif args.visualization == 'calibration_multiclass':
        calibration_multiclass(**vars(args))
    elif args.visualization == 'confusion_matrix':
        confusion_matrix(**vars(args))
    elif args.visualization == 'frequency_vs_f1':
        frequency_vs_f1(**vars(args))
    elif args.visualization == 'learning_curves':
        learning_curves(**vars(args))
    else:
        logging.info('Visualization argument not recognized')


if __name__ == '__main__':
    cli(sys.argv[1:])



# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging

import numpy as np
import tensorflow as tf

from ludwig.models.modules.initializer_modules import get_initializer
from ludwig.utils.data_utils import load_pretrained_embeddings


def embedding_matrix(vocab, embedding_size, representation='dense',
                     embeddings_trainable=True, pretrained_embeddings=None,
                     initializer=None, regularizer=None):
    vocab_size = len(vocab)
    if representation == 'dense':
        if pretrained_embeddings is not None and pretrained_embeddings is not False:
            embeddings_matrix = load_pretrained_embeddings(
                pretrained_embeddings, vocab)
            if embeddings_matrix.shape[-1] != embedding_size:
                raise ValueError(
                    'The size of the pretrained embedding size is {}, '
                    'but the specified embedding_size is {}. '
                    'Please change the embedding_size accordingly.'.format(
                        embeddings_matrix.shape[-1],
                        embedding_size
                    ))
            initializer_obj = tf.constant(embeddings_matrix, dtype=tf.float32)
        else:
            if vocab_size < embedding_size:
                logging.info(
                    '  embedding_size ({}) is greater than vocab_size ({}). '
                    'Setting embedding size to be equal to vocab_size.'.format(
                        embedding_size, vocab_size
                    ))
                embedding_size = vocab_size

            if initializer is not None:
                initializer_obj_ref = get_initializer(initializer)
            else:
                initializer_obj_ref = get_initializer(
                    {'type': 'uniform', 'minval': -1.0, 'maxval': 1.0})
            initializer_obj = initializer_obj_ref([vocab_size, embedding_size])

        embeddings = tf.get_variable('embeddings',
                                     initializer=initializer_obj,
                                     trainable=embeddings_trainable,
                                     regularizer=regularizer)

    elif representation == 'sparse':
        embedding_size = vocab_size
        embeddings = tf.get_variable('embeddings',
                                     initializer=get_initializer('identity')(
                                         [vocab_size, embedding_size]),
                                     trainable=False)

    else:
        raise Exception(
            'Embedding representation {} not supported.'.format(representation))

    return embeddings, embedding_size


class Embed:
    def __init__(
            self,
            vocab,
            embedding_size,
            representation='dense',
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            dropout=False,
            initializer=None,
            regularize=True
    ):
        self.vocab = vocab
        self.embedding_size = embedding_size
        self.representation = representation
        self.embeddings_trainable = embeddings_trainable
        self.pretrained_embeddings = pretrained_embeddings
        self.embeddings_on_cpu = embeddings_on_cpu
        self.dropout = dropout
        self.initializer = initializer
        self.regularize = regularize

    def __call__(
            self,
            input_ids,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        if not self.regularize:
            regularizer = None

        if self.embeddings_on_cpu:
            with tf.device('/cpu:0'):
                embeddings, embedding_size = embedding_matrix(
                    self.vocab,
                    self.embedding_size,
                    representation=self.representation,
                    embeddings_trainable=self.embeddings_trainable,
                    pretrained_embeddings=self.pretrained_embeddings,
                    initializer=self.initializer,
                    regularizer=regularizer
                )
        else:
            embeddings, embedding_size = embedding_matrix(
                self.vocab,
                self.embedding_size,
                representation=self.representation,
                embeddings_trainable=self.embeddings_trainable,
                pretrained_embeddings=self.pretrained_embeddings,
                initializer=self.initializer,
                regularizer=regularizer
            )
        logging.debug('  embeddings: {0}'.format(embeddings))

        embedded = tf.nn.embedding_lookup(embeddings, input_ids,
                                          name='embeddings_lookup')
        logging.debug('  embedded: {0}'.format(embedded))

        if self.dropout and dropout_rate is not None:
            embedded = tf.layers.dropout(embedded, rate=dropout_rate,
                                         training=is_training)
            logging.debug('  embedded_dropout: {}'.format(embedded))

        return embedded, embedding_size


class EmbedWeighted:
    def __init__(
            self,
            vocab,
            embedding_size,
            representation='dense',
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            dropout=False,
            initializer=None,
            regularize=True
    ):
        self.vocab = vocab
        self.embedding_size = embedding_size
        self.representation = representation
        self.embeddings_trainable = embeddings_trainable
        self.pretrained_embeddings = pretrained_embeddings
        self.embeddings_on_cpu = embeddings_on_cpu
        self.dropout = dropout
        self.initializer = initializer
        self.regularize = regularize

    def __call__(
            self,
            input_ids,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        if not self.regularize:
            regularizer = None

        if self.embeddings_on_cpu:
            with tf.device('/cpu:0'):
                embeddings, embedding_size = embedding_matrix(
                    self.vocab,
                    self.embedding_size,
                    representation=self.representation,
                    embeddings_trainable=self.embeddings_trainable,
                    pretrained_embeddings=self.pretrained_embeddings,
                    initializer=self.initializer,
                    regularizer=regularizer
                )
        else:
            embeddings, embedding_size = embedding_matrix(
                self.vocab,
                self.embedding_size,
                representation=self.representation,
                embeddings_trainable=self.embeddings_trainable,
                pretrained_embeddings=self.pretrained_embeddings,
                initializer=self.initializer,
                regularizer=regularizer
            )
        logging.debug('  embeddings: {0}'.format(embeddings))

        signed_input = tf.cast(tf.sign(tf.abs(input_ids)), tf.int32)
        multiple_hot_indexes = tf.multiply(
            signed_input,
            tf.constant(np.array([range(len(self.vocab))], dtype=np.int32))
        )
        embedded = tf.nn.embedding_lookup(
            embeddings,
            multiple_hot_indexes,
            name='embeddings_lookup'
        )
        logging.debug('  embedded: {0}'.format(embedded))

        # Get the multipliers to embeddings
        weights_mask = tf.expand_dims(input_ids, -1)
        weighted_embedded = tf.multiply(embedded, weights_mask)
        logging.debug('  weighted_embedded: {0}'.format(weighted_embedded))

        embedded_reduced = tf.reduce_sum(weighted_embedded, 1)
        logging.debug('  embedded_reduced: {0}'.format(embedded_reduced))

        if self.dropout and dropout_rate is not None:
            embedded = tf.layers.dropout(embedded, rate=dropout_rate,
                                         training=is_training)
            logging.debug('  embedded_dropout: {}'.format(embedded))

        return embedded_reduced, embedding_size


class EmbedSparse:
    def __init__(
            self,
            vocab,
            embedding_size,
            representation='dense',
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            reduce_output='sum',
            dropout=False,
            initializer=None,
            regularize=True
    ):
        self.vocab = vocab
        self.embedding_size = embedding_size
        self.representation = representation
        self.embeddings_trainable = embeddings_trainable
        self.pretrained_embeddings = pretrained_embeddings
        self.embeddings_on_cpu = embeddings_on_cpu
        self.reduce_output = reduce_output
        self.dropout = dropout
        self.initializer = initializer
        self.regularize = regularize

    def __call__(
            self,
            input_sparse,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        if not self.regularize:
            regularizer = None

        if self.embeddings_on_cpu:
            with tf.device('/cpu:0'):
                embeddings, embedding_size = embedding_matrix(
                    self.vocab,
                    self.embedding_size,
                    representation=self.representation,
                    embeddings_trainable=self.embeddings_trainable,
                    pretrained_embeddings=self.pretrained_embeddings,
                    initializer=self.initializer,
                    regularizer=regularizer
                )
        else:
            embeddings, embedding_size = embedding_matrix(
                self.vocab,
                self.embedding_size,
                representation=self.representation,
                embeddings_trainable=self.embeddings_trainable,
                pretrained_embeddings=self.pretrained_embeddings,
                initializer=self.initializer,
                regularizer=regularizer
            )
        logging.debug('  embeddings: {0}'.format(embeddings))

        multiple_hot_indexes = tf.multiply(
            input_sparse,
            tf.constant(np.array([range(len(self.vocab))], dtype=np.int32))
        )

        idx = tf.where(tf.not_equal(multiple_hot_indexes, 0))

        sparse_multiple_hot_indexes = tf.SparseTensor(
            idx,
            tf.gather_nd(multiple_hot_indexes, idx),
            tf.shape(multiple_hot_indexes, out_type=tf.int64)
        )

        embedded_reduced = tf.nn.embedding_lookup_sparse(
            embeddings,
            sparse_multiple_hot_indexes,
            sp_weights=None,
            combiner=self.reduce_output
        )
        logging.debug('  embedded_reduced: {0}'.format(embedded_reduced))

        # Old dense implementation
        # embedded = tf.nn.embedding_lookup(
        #     feature_embeddings,
        #     multiple_hot_indexes,
        #     name=input_feature['name'] + '_embeddings_lookup',
        # )
        # mask = tf.cast(tf.expand_dims(tf.sign(tf.abs(multiple_hot_indexes)), -1), tf.float32)
        # masked_embedded = tf.multiply(embedded, mask)
        # embedded_reduced = tf.reduce_sum(masked_embedded, 1)

        if self.dropout and dropout_rate is not None:
            embedded_reduced = tf.layers.dropout(embedded_reduced,
                                                 rate=dropout_rate,
                                                 training=is_training)
            logging.debug(
                '  embedded_reduced_dropout: {}'.format(embedded_reduced))

        return embedded_reduced, embedding_size


class EmbedSequence:
    def __init__(
            self,
            vocab,
            embedding_size,
            representation='dense',
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            mask=True,
            dropout=False,
            initializer=None,
            regularize=True
    ):
        self.embed = Embed(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            initializer=initializer,
            regularize=regularize
        )

        self.mask = mask

    def __call__(
            self,
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        embedded, embedding_size = self.embed(
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
        )

        if self.mask:
            mask_matrix = tf.cast(
                tf.expand_dims(tf.sign(tf.abs(input_sequence)), -1),
                dtype=tf.float32
            )
            embedded = tf.multiply(embedded, mask_matrix)

        return embedded, embedding_size



# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from collections import OrderedDict

import tensorflow as tf

from ludwig.features.feature_registries import output_type_registry
from ludwig.utils.algorithms_utils import topological_sort_feature_dependencies
from ludwig.utils.misc import get_from_registry


def build_outputs(output_features, hidden, hidden_size, regularizer,
                  dropout_rate,
                  is_training=True, **kwargs):
    output_features = topological_sort_feature_dependencies(output_features)
    output_tensors = OrderedDict()
    final_hidden = {}
    output_train_losses = []
    output_eval_losses = []
    for output_feature in output_features:
        of_train_mean_loss, of_eval_loss, of_output_tensors = build_single_output(
            output_feature,
            hidden,
            hidden_size,
            final_hidden,
            dropout_rate,
            regularizer,
            is_training,
            **kwargs
        )
        output_train_losses.append(of_train_mean_loss)
        output_eval_losses.append(of_eval_loss)
        output_tensors.update(of_output_tensors)

    train_combined_mean_loss = tf.reduce_sum(tf.stack(output_train_losses),
                                             name='train_combined_mean_loss')
    if regularizer is not None:
        regularization_losses = tf.get_collection(
            tf.GraphKeys.REGULARIZATION_LOSSES)
        if regularization_losses:
            regularization_loss = tf.add_n(regularization_losses)
            logging.debug('- Regularization losses: {0}'.format(
                regularization_losses))

        else:
            regularization_loss = tf.constant(0.0)
    else:
        regularization_loss = tf.constant(0.0)

    train_reg_mean_loss = tf.add(train_combined_mean_loss, regularization_loss,
                                 name='train_combined_regularized_mean_loss')

    eval_combined_loss = tf.reduce_sum(tf.stack(output_eval_losses), axis=0,
                                       name='eval_combined_loss')

    return train_reg_mean_loss, eval_combined_loss, regularization_loss, output_tensors


def build_single_output(output_feature, feature_hidden, feature_hidden_size,
                        final_hidden,
                        dropout_rate, regularizer, is_training=True, **kwargs):
    logging.debug('- Output {} feature {}'.format(
        output_feature['type'],
        output_feature['name']
    ))
    with tf.variable_scope(output_feature['name']):
        feature_class = get_from_registry(
            output_feature['type'],
            output_type_registry
        )
        feature = feature_class(output_feature)
        weighted_train_mean_loss, weighted_eval_loss, output_tensors = feature.concat_dependencies_and_build_output(
            feature_hidden,
            feature_hidden_size,
            final_hidden,
            dropout_rate=dropout_rate,
            regularizer=regularizer,
            is_training=is_training,
            **kwargs
        )
    return weighted_train_mean_loss, weighted_eval_loss, output_tensors



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from math import floor, ceil

import numpy as np
from skimage.transform import resize

from ludwig.constants import CROP_OR_PAD, INTERPOLATE


def pad(img, size, axis):
    old_size = img.shape[axis]
    pad_size = float(size - old_size) / 2
    pads = [(0, 0), (0, 0), (0, 0)]
    pads[axis] = (floor(pad_size), ceil(pad_size))
    return np.pad(img, pads, 'edge')


def crop(img, size, axis):
    y_min = 0
    y_max = img.shape[0]
    x_min = 0
    x_max = img.shape[1]
    if axis == 0:
        y_min = int(float(y_max - size) / 2)
        y_max = y_min + size
    else:
        x_min = int(float(x_max - size) / 2)
        x_max = x_min + size

    return img[y_min: y_max, x_min: x_max, :]


def crop_or_pad(img, new_size_tuple):
    for axis in range(2):
        if new_size_tuple[axis] != img.shape[axis]:
            if new_size_tuple[axis] > img.shape[axis]:
                img = pad(img, new_size_tuple[axis], axis)
            else:
                img = crop(img, new_size_tuple[axis], axis)
    return img


def resize_image(img, new_size_typle, resize_method):
    if resize_method == CROP_OR_PAD:
        return crop_or_pad(img, new_size_typle)
    elif resize_method == INTERPOLATE:
        return resize(img, new_size_typle)
    raise ValueError('Invalid image resize method: {}'.format(resize_method))



# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging

import tensorflow as tf


def reduce_feed_forward_attention(current_inputs, hidden_size=256):
    with tf.variable_scope('reduce_ff_attention'):
        weights_1 = tf.get_variable('weights_1',
                                    [current_inputs.shape[-1], hidden_size])
        logging.debug('  att_weights_1: {}'.format(weights_1))
        biases_1 = tf.get_variable('biases_1', [hidden_size])
        logging.debug('  att_biases_1: {}'.format(biases_1))
        weights_2 = tf.get_variable('weights_2', [hidden_size, 1])
        logging.debug('  att_weights_2: {}'.format(weights_2))

        current_inputs_reshape = tf.reshape(current_inputs,
                                            [-1, current_inputs.shape[-1]])
        hidden = tf.tanh(
            tf.matmul(current_inputs_reshape, weights_1) + biases_1)
        logging.debug('  att_hidden: {}'.format(hidden))
        attention = tf.nn.softmax(tf.reshape(tf.matmul(hidden, weights_2),
                                             [-1, tf.shape(current_inputs)[1]]))
        logging.debug('  att_attention: {}'.format(attention))
        # attention [bs x seq]
        geated_inputs = tf.reduce_sum(
            tf.expand_dims(attention, -1) * current_inputs, 1)
        logging.debug('  att_geated_inputs: {}'.format(geated_inputs))
    return geated_inputs


def feed_forward_attention(current_inputs, feature_hidden_size,
                           hidden_size=256):
    with tf.variable_scope('ff_attention'):
        geated_inputs = reduce_feed_forward_attention(current_inputs,
                                                      hidden_size=hidden_size)

        # stacking inputs and attention vectors
        tiled_geated_inputs = tf.tile(tf.expand_dims(geated_inputs, 1),
                                      [1, tf.shape(current_inputs)[1], 1])
        logging.debug(
            '  att_tiled_geated_inputs: {}'.format(tiled_geated_inputs))
        outputs = tf.concat([current_inputs, tiled_geated_inputs],
                            axis=-1)  # [bs x s1 x 2*h]
        logging.debug('  att_outputs: {}'.format(outputs))
        # outputs = current_inputs + context # [bs x s1 x h]

    return outputs, feature_hidden_size * 2


def simple_memory_attention(current_inputs, context):
    assert current_inputs.shape[2] == context.shape[2]
    # calculating attention
    attention = tf.nn.softmax(
        tf.matmul(current_inputs, context, transpose_b=True))  # [bs x s1 x s2]
    logging.debug('  att_outputs: {}'.format(attention))

    # weighted_sum(attention, encoding_sequence_embedded)
    exp_ese = tf.expand_dims(context, 1)  # [bs x 1 x s2 x h]
    exp_att = tf.expand_dims(attention, -1)  # [bs x s1 x s2 x 1]
    weighted_sum = tf.multiply(exp_ese, exp_att)  # [bs x s1 x s2 x h]
    reduced_weighted_sum = tf.reduce_sum(weighted_sum, axis=2)  # [bs x s1 x h]
    logging.debug('  att_reduced_weighted_sum: {}'.format(reduced_weighted_sum))

    # stacking inputs and attention vectors
    outputs = tf.concat([current_inputs, reduced_weighted_sum],
                        axis=-1)  # [bs x s1 x 2*h]
    logging.debug('  att_outputs: {}'.format(outputs))

    return outputs, outputs.shape[-1]


def feed_forward_memory_attention(current_inputs, memory, hidden_size=256):
    seq_len = tf.shape(current_inputs)[1]
    mem_len = tf.shape(current_inputs)[1]
    seq_width = current_inputs.shape[2]
    mem_width = memory.shape[2]

    inputs_tile = tf.reshape(tf.tile(current_inputs, [1, 1, mem_len]),
                             [-1, seq_len, mem_len, seq_width])
    context_tile = tf.reshape(tf.tile(memory, [1, seq_len, 1]),
                              [-1, seq_len, mem_len, mem_width])
    concat_tile = tf.concat([inputs_tile, context_tile],
                            axis=-1)  # [bs, seq, seq, seq_w + ctx_w]
    logging.debug('  att_input_context_concat: {}'.format(concat_tile))

    with tf.variable_scope('reduce_contextual_ff_attention'):
        weights_1 = tf.get_variable('weights_1',
                                    [concat_tile.shape[-1], hidden_size])
        logging.debug('  att_weights_1: {}'.format(weights_1))
        biases_1 = tf.get_variable('biases_1', [hidden_size])
        logging.debug('  att_biases_1: {}'.format(biases_1))
        weights_2 = tf.get_variable('weights_2', [hidden_size, 1])
        logging.debug('  att_weights_2: {}'.format(weights_2))

        current_inputs_reshape = tf.reshape(concat_tile,
                                            [-1, concat_tile.shape[-1]])
        hidden = tf.tanh(
            tf.matmul(current_inputs_reshape, weights_1) + biases_1)
        logging.debug('  att_hidden: {}'.format(hidden))
        attention = tf.nn.softmax(
            tf.reshape(tf.matmul(hidden, weights_2), [-1, seq_len, mem_len]))
        logging.debug('  att_attention: {}'.format(attention))
        # attention [bs x seq]
        geated_inputs = tf.reduce_sum(
            tf.expand_dims(attention, -1) * inputs_tile, 2)
        logging.debug('  att_geated_inputs: {}'.format(geated_inputs))

    return geated_inputs, geated_inputs.shape.as_list()[-1]



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import logging
import os
import sys

import numpy as np

from ludwig.data.preprocessing import preprocess_for_prediction
from ludwig.globals import LUDWIG_VERSION
from ludwig.globals import TRAIN_SET_METADATA_FILE_NAME
from ludwig.models.model import load_model_and_definition
from ludwig.utils.print_utils import logging_level_registry
from ludwig.utils.print_utils import print_boxed
from ludwig.utils.print_utils import print_ludwig
from ludwig.utils.strings_utils import make_safe_filename


def collect_activations(
        model_path,
        tensors,
        data_csv=None,
        data_hdf5=None,
        dataset_type='generic',
        split='test',
        batch_size=128,
        output_directory='results',
        gpus=None,
        gpu_fraction=1.0,
        debug=False,
        **kwargs
):
    """Uses the pretrained model to collect the tensors corresponding to a
    datapoint in the dataset. Saves the tensors to the experiment directory

    :param model_path: Is the model from which the tensors will be collected
    :param tensors: List contaning the names of the tensors to collect
    :param data_csv: The CSV filepath which contains the datapoints from which
           the tensors are collected
    :param data_hdf5: The HDF5 file path if the CSV file path does not exist,
           an alternative source of providing the data to the model
    :param dataset_type: Dataset type
    :param split: Split type
    :param batch_size: Batch size
    :param output_directory: Output directory
    :param gpus: The total number of GPUs that the model intends to use
    :param gpu_fraction: The fraction of each GPU that the model intends on
           using
    :param debug: To step through the stack traces and find possible errors
    :returns: None

    """
    # setup directories and file names
    experiment_dir_name = output_directory
    suffix = 0
    while os.path.exists(experiment_dir_name):
        experiment_dir_name = output_directory + '_' + str(suffix)
        suffix += 1

    logging.info('Dataset type: {}'.format(dataset_type))
    logging.info('Dataset path: {}'.format(
        data_csv if data_csv is not None else data_hdf5)
    )
    logging.info('Model path: {}'.format(model_path))
    logging.info('Output path: {}'.format(experiment_dir_name))
    logging.info('\n')

    train_set_metadata_fp = os.path.join(
        model_path,
        TRAIN_SET_METADATA_FILE_NAME
    )

    # preprocessing
    dataset, train_set_metadata = preprocess_for_prediction(
        model_path,
        split,
        dataset_type,
        data_csv,
        data_hdf5,
        train_set_metadata_fp
    )

    model, model_definition = load_model_and_definition(model_path)

    # collect activations
    print_boxed('COLLECT ACTIVATIONS')
    collected_tensors = model.collect_activations(
        dataset,
        tensors,
        batch_size,
        gpus=gpus,
        gpu_fraction=gpu_fraction
    )

    model.close_session()

    # saving
    os.mkdir(experiment_dir_name)
    save_tensors(collected_tensors, experiment_dir_name)

    logging.info('Saved to: {0}'.format(experiment_dir_name))


def collect_weights(
        model_path,
        tensors,
        output_directory='results',
        debug=False,
        **kwargs
):
    # setup directories and file names
    experiment_dir_name = output_directory
    suffix = 0
    while os.path.exists(experiment_dir_name):
        experiment_dir_name = output_directory + '_' + str(suffix)
        suffix += 1

    logging.info('Model path: {}'.format(model_path))
    logging.info('Output path: {}'.format(experiment_dir_name))
    logging.info('\n')

    model, model_definition = load_model_and_definition(model_path)

    # collect weights
    print_boxed('COLLECT WEIGHTS')
    collected_tensors = model.collect_weights(tensors)
    model.close_session()

    # saving
    os.mkdir(experiment_dir_name)
    save_tensors(collected_tensors, experiment_dir_name)

    logging.info('Saved to: {0}'.format(experiment_dir_name))


def save_tensors(collected_tensors, experiment_dir_name):
    for tensor_name, tensor_values in collected_tensors.items():
        np_filename = os.path.join(
            experiment_dir_name,
            make_safe_filename(tensor_name) + '.npy'
        )
        np.save(np_filename, tensor_values)


def cli_collect_activations(sys_argv):
    """Command Line Interface to communicate with the collection of tensors and
    there are several options that can specified when calling this function:

    --data_csv: Filepath for the input csv
    --data_hdf5: Filepath for the input hdf5 file, if there is a csv file, this
                 is not read
    --d: Refers to the dataset type of the file being read, by default is
         *generic*
    --s: Refers to the split of the data, can be one of: train, test,
         validation, full
    --m: Input model that is necessary to collect to the tensors, this is a
         required *option*
    --t: Tensors to collect
    --od: Output directory of the model, defaults to results
    --bs: Batch size
    --g: Number of gpus that are to be used
    --gf: Fraction of each GPUs memory to use.
    --dbg: Debug if the model is to be started with python debugger
    --v: Verbose: Defines the logging level that the user will be exposed to
    """
    parser = argparse.ArgumentParser(
        description='This script loads a pretrained model and uses it collect '
                    'tensors for each datapoint in the dataset.',
        prog='ludwig collect_activations',
        usage='%(prog)s [options]')

    # ---------------
    # Data parameters
    # ---------------
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--data_csv', help='input data CSV file')
    group.add_argument('--data_hdf5', help='input data HDF5 file')

    parser.add_argument(
        '-s',
        '--split',
        default='test',
        choices=['training', 'validation', 'test', 'full'],
        help='the split to test the model on'
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument(
        '-m',
        '--model_path',
        help='model to load',
        required=True
    )
    parser.add_argument(
        '-t',
        '--tensors',
        help='tensors to collect',
        nargs='+',
        required=True
    )

    # -------------------------
    # Output results parameters
    # -------------------------
    parser.add_argument(
        '-od',
        '--output_directory',
        type=str,
        default='results',
        help='directory that contains the results'
    )

    # ------------------
    # Generic parameters
    # ------------------
    parser.add_argument(
        '-bs',
        '--batch_size',
        type=int,
        default=128,
        help='size of batches'
    )

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        '-g',
        '--gpus',
        type=int,
        default=0,
        help='list of gpu to use'
    )
    parser.add_argument(
        '-gf',
        '--gpu_fraction',
        type=float,
        default=1.0,
        help='fraction of gpu memory to initialize the process with'
    )
    parser.add_argument(
        '-dbg',
        '--debug',
        action='store_true',
        default=False,
        help='enables debugging mode'
    )
    parser.add_argument(
        '-l',
        '--logging_level',
        default='info',
        help='the level of logging to use',
        choices=['critical', 'error', 'warning', 'info', 'debug', 'notset']
    )

    args = parser.parse_args(sys_argv)

    logging.basicConfig(
        stream=sys.stdout,
        level=logging_level_registry[args.logging_level],
        format='%(message)s'
    )

    print_ludwig('Collect Activations', LUDWIG_VERSION)

    collect_activations(**vars(args))


def cli_collect_weights(sys_argv):
    """Command Line Interface to collecting the weights for the model
    --m: Input model that is necessary to collect to the tensors, this is a
         required *option*
    --t: Tensors to collect
    --od: Output directory of the model, defaults to results
    --dbg: Debug if the model is to be started with python debugger
    --v: Verbose: Defines the logging level that the user will be exposed to
    """
    parser = argparse.ArgumentParser(
        description='This script loads a pretrained model '
                    'and uses it collect weights.',
        prog='ludwig collect_weights',
        usage='%(prog)s [options]'
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument(
        '-m',
        '--model_path',
        help='model to load',
        required=True
    )
    parser.add_argument(
        '-t',
        '--tensors',
        help='tensors to collect',
        nargs='+',
        required=True
    )

    # -------------------------
    # Output results parameters
    # -------------------------
    parser.add_argument(
        '-od',
        '--output_directory',
        type=str,
        default='results',
        help='directory that contains the results'
    )

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        '-dbg',
        '--debug',
        action='store_true',
        default=False,
        help='enables debugging mode'
    )
    parser.add_argument(
        '-l',
        '--logging_level',
        default='info',
        help='the level of logging to use',
        choices=['critical', 'error', 'warning', 'info', 'debug', 'notset']
    )

    args = parser.parse_args(sys_argv)

    logging.basicConfig(
        stream=sys.stdout,
        level=logging_level_registry[args.logging_level],
        format='%(message)s'
    )

    print_ludwig('Collect Weights', LUDWIG_VERSION)
    collect_weights(**vars(args))


if __name__ == '__main__':
    if len(sys.argv) > 1:
        if sys.argv[1] == 'activations':
            cli_collect_activations(sys.argv[2:])
        elif sys.argv[1] == 'weights':
            cli_collect_weights(sys.argv[2:])
        else:
            print('Unrecognized command')
    else:
        print('Unrecognized command')



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import absolute_import
from __future__ import division

import copy
import logging
from collections import Counter

import matplotlib as mpl
from matplotlib import ticker
from matplotlib.lines import Line2D

mpl.use('TkAgg')  # make matplotlib run on mac os
import matplotlib.patches as patches
import matplotlib.path as path
import matplotlib.patheffects as PathEffects
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from mpl_toolkits.mplot3d import Axes3D


# plt.rc('xtick', labelsize='x-large')
# plt.rc('ytick', labelsize='x-large')
# plt.rc('axes', labelsize='x-large')

def lerning_curves_plot(train_values, vali_values, metric, algorithm_names=None,
                        title=None):
    num_algorithms = len(train_values)
    max_len = max([len(tv) for tv in train_values])

    fig, ax = plt.subplots()

    sns.set_style('whitegrid')

    if title is not None:
        ax.set_title(title)

    if num_algorithms == 1:
        colors = plt.get_cmap('tab10').colors
    else:  # num_algorithms > 1
        colors = plt.get_cmap('tab20').colors

    ax.grid(which='both')
    ax.grid(which='minor', alpha=0.5)
    ax.grid(which='major', alpha=0.75)
    ax.set_xlabel('epochs')
    ax.set_ylabel(metric.replace('_', ' '))

    xs = list(range(1, max_len + 1))

    for i in range(num_algorithms):
        name_prefix = algorithm_names[
                          i] + ' ' if algorithm_names is not None and i < len(
            algorithm_names) else ''
        ax.plot(xs, train_values[i], label=name_prefix + 'training',
                color=colors[i * 2], linewidth=3)
        if i < len(vali_values):
            ax.plot(xs, vali_values[i], label=name_prefix + 'validation',
                    color=colors[i * 2 + 1], linewidth=3)

    ax.legend()
    plt.tight_layout()
    plt.show()


def compare_classifiers_plot(scores, metrics, algoritm_names=None,
                             adaptive=False, decimals=4, title=None):
    assert len(scores) == len(metrics)
    assert len(scores) > 0

    num_metrics = len(metrics)

    sns.set_style('whitegrid')

    fig, ax = plt.subplots()

    ax.grid(which='both')
    ax.grid(which='minor', alpha=0.5)
    ax.grid(which='major', alpha=0.75)
    ax.set_xticklabels([], minor=True)

    if title is not None:
        ax.set_title(title)

    width = 0.8 / num_metrics if num_metrics > 1 else 0.4
    ticks = np.arange(len(scores[0]))

    colors = plt.get_cmap('tab10').colors
    if adaptive:
        maximum = max([max(score) for score in scores])
    else:
        ax.set_xlim([0, 1])
        ax.set_xticks(np.linspace(0.0, 1.0, num=21), minor=True)
        ax.set_xticks(np.linspace(0.0, 1.0, num=11))
        maximum = 1

    half_total_width = 0.4 if num_metrics > 1 else 0.2
    ax.set_yticks(ticks + half_total_width - width / 2)
    ax.set_yticklabels(algoritm_names if algoritm_names is not None else '')
    ax.invert_yaxis()  # labels read top-to-bottom

    for i, metric in enumerate(metrics):
        ax.barh(ticks + (i * width), scores[i], width, label=metric,
                color=colors[i])

        for j, v in enumerate(scores[i]):
            if v < maximum * (0.025 * decimals + 0.1):
                x = v + maximum * 0.01
                horizontal_alignment = 'left'
            else:
                x = v - maximum * 0.01
                horizontal_alignment = 'right'
            txt = ax.text(x, ticks[j] + (i * width),
                          ('{:.' + str(decimals) + 'f}').format(v),
                          color='white',
                          fontweight='bold', verticalalignment='center',
                          horizontalalignment=horizontal_alignment)
            txt.set_path_effects(
                [PathEffects.withStroke(linewidth=3, foreground='black')])

    plt.setp(ax.get_xminorticklabels(), visible=False)

    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.tight_layout()
    plt.show()


def compare_classifiers_line_plot(xs, scores, metric, algorithm_names=None,
                                  title=None):
    sns.set_style('whitegrid')
    colors = plt.get_cmap('tab10').colors

    fig, ax = plt.subplots()

    ax.grid(which='both')
    ax.grid(which='minor', alpha=0.5)
    ax.grid(which='major', alpha=0.75)

    if title is not None:
        ax.set_title(title)

    ax.set_xticks(xs)
    ax.set_xticklabels(xs)
    ax.set_xlabel('k')
    ax.set_ylabel(metric)

    for i, score in enumerate(scores):
        ax.plot(xs, score,
                label=algorithm_names[
                    i] if algorithm_names is not None and i < len(
                    algorithm_names) else 'Algorithm {}'.format(i),
                color=colors[i], linewidth=3, marker='o')

    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.tight_layout()
    plt.show()


def compare_classifiers_multiclass_multimetric_plot(scores, metrics,
                                                    labels=None, title=None):
    assert len(scores) > 0

    sns.set_style('whitegrid')

    fig, ax = plt.subplots()

    if title is not None:
        ax.set_title(title)

    width = 0.9 / len(scores)
    ticks = np.arange(len(scores[0]))

    colors = plt.get_cmap('tab10').colors
    ax.set_xlabel('class')
    ax.set_xticks(ticks + width)
    if labels is not None:
        ax.set_xticklabels(labels, rotation=90)
    else:
        ax.set_xticklabels(ticks, rotation=90)

    for i, score in enumerate(scores):
        ax.bar(ticks + i * width, score, width, label=metrics[i],
               color=colors[i])

    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.tight_layout()
    plt.show()


def radar_chart(ground_truth, predictions, algorithms=None, log_scale=False,
                title=None):
    sns.set_style('whitegrid')

    if title is not None:
        plt.title(title)

    ground_truth = ground_truth[0:10]
    predictions = [pred[0:10] for pred in predictions]

    gt_argsort = np.argsort(-ground_truth)  # sort deacreasing
    logging.info(gt_argsort)
    ground_truth = ground_truth[gt_argsort]
    predictions = [pred[gt_argsort] for pred in predictions]

    maximum = max(max(ground_truth), max([max(p) for p in predictions]))

    ax = plt.subplot(111, polar=True)
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    ax.set_rmax(maximum)
    ax.set_rlabel_position(305)
    ax.set_ylabel('Probability')
    # ax.set_rscale('log')
    ax.grid(True)

    colors = plt.get_cmap('tab10').colors

    num_classes = len(ground_truth)

    # Set ticks to the number of properties (in radians)
    t = np.arange(0, 2 * np.pi, 2 * np.pi / num_classes)
    ax.set_xticks(t, [])
    ax.set_xticklabels(np.arange(0, num_classes))

    # Set yticks from 0 to 10
    # ax.set_yticks(np.linspace(0, 10, 11))
    # Set axes limits
    # ax.set_rlim(0, 1)
    # ax.set_rscale('log')

    def draw_polygon(values, label, color='grey'):
        points = [(x, y) for x, y in zip(t, values)]
        points.append(points[0])
        points = np.array(points)

        codes = [path.Path.MOVETO, ] + \
                [path.Path.LINETO, ] * (len(values) - 1) + \
                [path.Path.CLOSEPOLY]
        _path = path.Path(points, codes)
        _patch = patches.PathPatch(_path, fill=True, color=color, linewidth=0,
                                   alpha=.2)
        ax.add_patch(_patch)
        _patch = patches.PathPatch(_path, fill=False, color=color, linewidth=3)
        ax.add_patch(_patch)

        # Draw circles at value points
        # line = ax.scatter(points[:, 0], points[:, 1], linewidth=3,
        #            s=50, color='white', edgecolor=color, zorder=10)
        ax.plot(points[:, 0], points[:, 1], linewidth=3, marker='o',
                fillstyle='full',
                markerfacecolor='white',
                markeredgecolor=color,
                markeredgewidth=2,
                color=color, zorder=10, label=label)

    draw_polygon(ground_truth, 'Ground Truth')

    # Draw polygon representing values
    for i, alg_predictions in enumerate(predictions):
        draw_polygon(alg_predictions, algorithms[i], colors[i])

    ax.legend(frameon=True, loc='upper left')
    plt.tight_layout()
    plt.show()


def pie(ax, values, **kwargs):
    total = sum(values)

    def formatter(pct):
        if pct > 0:
            return '{:0.0f}\n({:0.1f}%)'.format(pct * total / 100, pct)
        else:
            return ''

    wedges, _, labels = ax.pie(values, autopct=formatter, **kwargs)
    return wedges


def donut(inside_values, inside_labels, outside_values, outside_labels,
          outside_groups, title=None):
    fig, ax = plt.subplots()

    if title is not None:
        ax.set_title(title)

    ax.axis('equal')

    width = 0.35
    colors_tab20c = list(plt.get_cmap('tab20c').colors)
    colors_set2 = list(plt.get_cmap('Set2').colors)
    colors_set3 = list(plt.get_cmap('Set3').colors)
    colors_pastel1 = list(plt.get_cmap('Pastel1').colors)

    # swap green and red
    # for i in range(4):
    #    tmp = colors[4 + i]
    #    colors[4 + i] = colors[8 + i]
    #    colors[8 + i] = tmp

    colors = []
    colors.extend(colors_tab20c[8:12])
    colors.append(colors_set2[5])
    colors.append(colors_set3[11])
    colors.append(colors_set3[1])
    colors.append(colors_pastel1[5])
    colors.extend(colors_tab20c[4:8])

    inside_colors = [colors[x * 4] for x in range(len(inside_values))]

    group_count = Counter(outside_groups)
    outside_colors = [colors[(i * 4) + ((j % 3) + 1)]
                      for i in list(set(outside_groups))
                      for j in range(group_count[i])]

    outside = pie(ax, outside_values, radius=1, pctdistance=1 - width / 2,
                  colors=outside_colors, startangle=90, counterclock=False,
                  textprops={'color': 'w', 'weight': 'bold',
                             'path_effects': [
                                 PathEffects.withStroke(linewidth=3,
                                                        foreground='black')]})
    inside = pie(ax, inside_values, radius=1 - width,
                 pctdistance=1 - (width / 2) / (1 - width),
                 colors=inside_colors, startangle=90, counterclock=False,
                 textprops={'color': 'w', 'weight': 'bold',
                            'path_effects': [PathEffects.withStroke(linewidth=3,
                                                                    foreground='black')]})
    plt.setp(inside + outside, width=width, edgecolor='white')

    wedges = []
    labels = []
    so_far = 0
    for i in list(set(outside_groups)):
        wedges.append(inside[i])
        labels.append(inside_labels[i])
        for j in range(group_count[i]):
            wedges.append(outside[so_far])
            labels.append(outside_labels[so_far])
            so_far += 1

    ax.legend(wedges, labels, frameon=True)
    plt.tight_layout()
    plt.show()


def confidence_fitlering_plot(thresholds, accuracies, dataset_kepts,
                              algorithm_names=None, title=None):
    assert len(accuracies) == len(dataset_kepts)
    num_algorithms = len(accuracies)

    sns.set_style('whitegrid')

    if num_algorithms == 1:
        colors = plt.get_cmap('tab10').colors
    else:  # num_algorithms > 1
        colors = plt.get_cmap('tab20').colors

    y_ticks_minor = np.linspace(0.0, 1.0, num=21)
    y_ticks_major = np.linspace(0.0, 1.0, num=11)
    y_ticks_major_labels = ['{:3.0f}%'.format(y * 100) for y in y_ticks_major]

    fig, ax1 = plt.subplots()

    if title is not None:
        ax1.set_title(title)

    ax1.grid(which='both')
    ax1.grid(which='minor', alpha=0.5)
    ax1.grid(which='major', alpha=0.75)
    ax1.set_xticks([x for idx, x in enumerate(thresholds) if idx % 2 == 0])
    ax1.set_xticks(thresholds, minor=True)

    ax1.set_xlim(-0.05, 1.05)
    ax1.set_xlabel('confidence threshold')

    ax1.set_ylim(0, 1.05)
    ax1.set_yticks(y_ticks_major)
    ax1.set_yticklabels(y_ticks_major_labels)
    ax1.set_yticks(y_ticks_minor, minor=True)

    ax2 = ax1.twinx()

    ax2.set_ylim(0, 1.05)
    ax2.set_yticks(y_ticks_major)
    ax2.set_yticklabels(y_ticks_major_labels)
    ax2.set_yticks(y_ticks_minor, minor=True)

    for i in range(len(accuracies)):
        algorithm_name = algorithm_names[
                             i] + ' ' if algorithm_names is not None and i < len(
            algorithm_names) else ''
        ax1.plot(thresholds, accuracies[i],
                 label='{} accuracy'.format(algorithm_name),
                 color=colors[i * 2],
                 linewidth=3)
        ax1.plot(thresholds, dataset_kepts[i],
                 label='{} data coverage'.format(algorithm_name),
                 color=colors[i * 2 + 1], linewidth=3)

    ax1.legend(frameon=True, loc=3)
    plt.tight_layout()
    plt.show()


def confidence_fitlering_data_vs_acc_plot(accuracies, dataset_kepts,
                                          model_names=None,
                                          dotted=False,
                                          decimal_digits=0,
                                          y_label='accuracy',
                                          title=None):
    assert len(accuracies) == len(dataset_kepts)

    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab10').colors

    max_dataset_kept = max(
        [max(dataset_kept) for dataset_kept in dataset_kepts])

    x_ticks_minor = np.linspace(0.0, max_dataset_kept, num=21)
    x_ticks_major = np.linspace(0.0, max_dataset_kept, num=11)
    x_ticks_major_labels = [
        '{value:3.{decimal_digits}f}%'.format(
            decimal_digits=decimal_digits,
            value=x * 100
        ) for x in x_ticks_major
    ]
    y_ticks_minor = np.linspace(0.0, 1.0, num=21)
    y_ticks_major = np.linspace(0.0, 1.0, num=11)

    fig, ax = plt.subplots()

    if title is not None:
        ax.set_title(title)

    ax.grid(which='both')
    ax.grid(which='minor', alpha=0.5)
    ax.grid(which='major', alpha=0.75)
    ax.set_xticks(x_ticks_major)
    ax.set_xticks(x_ticks_minor, minor=True)
    ax.set_xticklabels(x_ticks_major_labels)
    ax.set_xlim(0, max_dataset_kept)
    ax.set_xlabel('data coverage')

    ax.set_ylim(0, 1)
    ax.set_yticks(y_ticks_major)
    ax.set_yticks(y_ticks_minor, minor=True)
    ax.set_ylabel(y_label)

    for i in range(len(accuracies)):
        curr_dotted = dotted[i] if isinstance(dotted,
                                              (list, tuple)) and i < len(
            dotted) else dotted
        algorithm_name = model_names[
                             i] + ' ' if model_names is not None and i < len(
            model_names) else ''
        ax.plot(dataset_kepts[i], accuracies[i], label=algorithm_name,
                color=colors[i],
                linewidth=3, linestyle=':' if curr_dotted else '-')

    ax.legend(frameon=True, loc=3)
    plt.tight_layout()
    plt.show()


def confidence_fitlering_data_vs_acc_multiline_plot(accuracies, dataset_kepts,
                                                    models_names,
                                                    title=None):
    assert len(accuracies) == len(dataset_kepts)

    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab20').colors

    max_dataset_kept = max(
        [max(dataset_kept) for dataset_kept in dataset_kepts])

    x_ticks_minor = np.linspace(0.0, max_dataset_kept, num=21)
    x_ticks_major = np.linspace(0.0, max_dataset_kept, num=11)
    x_ticks_major_labels = ['{:3.0f}%'.format(x * 100) for x in x_ticks_major]
    y_ticks_minor = np.linspace(0.0, 1.0, num=21)
    y_ticks_major = np.linspace(0.0, 1.0, num=11)

    fig, ax = plt.subplots()

    if title is not None:
        ax.set_title(title)

    ax.grid(which='both')
    ax.grid(which='minor', alpha=0.5)
    ax.grid(which='major', alpha=0.75)
    ax.set_xticks(x_ticks_major)
    ax.set_xticks(x_ticks_minor, minor=True)
    ax.set_xticklabels(x_ticks_major_labels)
    ax.set_xlim(0, max_dataset_kept)
    ax.set_xlabel('data coverage')

    ax.set_ylim(0, 1)
    ax.set_yticks(y_ticks_major)
    ax.set_yticks(y_ticks_minor, minor=True)
    ax.set_ylabel('accuracy')

    for i in range(len(accuracies)):
        ax.plot(dataset_kepts[i], accuracies[i], color=colors[0],
                linewidth=1.0, alpha=0.35)

    legend_elements = [Line2D([0], [0], linewidth=1.0, color=colors[0])]
    ax.legend(legend_elements, models_names)
    plt.tight_layout()
    plt.show()


def confidence_fitlering_3d_plot(thresholds_1, thresholds_2, accuracies,
                                 dataset_kepts, threshold_fields=None,
                                 title=None):
    assert len(accuracies) == len(dataset_kepts)
    assert len(thresholds_1) == len(thresholds_2)

    thresholds_1, thresholds_2 = np.meshgrid(thresholds_1, thresholds_2)

    colors = plt.get_cmap('tab10').colors
    sns.set_style('white')

    z_ticks_minor = np.linspace(0.0, 1.0, num=21)
    z_ticks_major = np.linspace(0.0, 1.0, num=11)
    z_ticks_major_labels = ['{:3.0f}%'.format(z * 100) for z in z_ticks_major]

    fig = plt.figure()
    ax = Axes3D
    ax = fig.add_subplot(111, projection='3d')

    if title is not None:
        ax.set_title(title)

    ax.grid(which='both')
    ax.grid(which='minor', alpha=0.5)
    ax.grid(which='major', alpha=0.75)

    ax.set_xlabel('{} probability'.format(threshold_fields[0]))
    ax.set_ylabel('{} probability'.format(threshold_fields[1]))

    ax.set_xlim(np.min(thresholds_1), np.max(thresholds_1))
    ax.set_ylim(np.min(thresholds_2), np.max(thresholds_2))
    ax.set_zlim(0, 1)
    ax.set_zticks(z_ticks_major)
    ax.set_zticklabels(z_ticks_major_labels)
    ax.set_zticks(z_ticks_minor, minor=True)

    # ORRIBLE HACK, IT'S THE ONLY WAY TO REMOVE PADDING
    from mpl_toolkits.mplot3d.axis3d import Axis
    if not hasattr(Axis, '_get_coord_info_old'):
        def _get_coord_info_new(self, renderer):
            mins, maxs, centers, deltas, tc, highs = self._get_coord_info_old(
                renderer)
            mins += deltas / 4
            maxs -= deltas / 4
            return mins, maxs, centers, deltas, tc, highs

        Axis._get_coord_info_old = Axis._get_coord_info
        Axis._get_coord_info = _get_coord_info_new
    # END OF HORRIBLE HACK

    surf_1 = ax.plot_surface(thresholds_1, thresholds_2, accuracies,
                             alpha=0.5,
                             label='accuracy',
                             cmap=plt.get_cmap('winter'),
                             edgecolor='none')
    surf_2 = ax.plot_surface(thresholds_1, thresholds_2, dataset_kepts,
                             alpha=0.5,
                             label='data coverage',
                             cmap=plt.get_cmap('autumn'),
                             edgecolor='none')

    handle_1 = copy.copy(surf_1)
    handle_2 = copy.copy(surf_2)

    handle_1.set_color(colors[0])
    handle_2.set_color(colors[1])

    handle_1._edgecolors2d = handle_1._edgecolors3d
    handle_2._edgecolors2d = handle_2._edgecolors3d

    handle_1._facecolors2d = handle_1._facecolors3d
    handle_2._facecolors2d = handle_2._facecolors3d

    ax.legend(frameon=True, loc=3, handles=[handle_1, handle_2])

    plt.tight_layout()
    plt.show()


def threshold_vs_metric_plot(thresholds, scores, algorithm_names=None,
                             title=None):
    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab10').colors

    # y_ticks_minor = np.linspace(0.0, 1.0, num=21)
    # y_ticks_major = np.linspace(0.0, 1.0, num=11)
    # y_ticks_major_labels = ['{:3.0f}%'.format(y * 100) for y in y_ticks_major]

    fig, ax1 = plt.subplots()

    if title is not None:
        ax1.set_title(title)

    ax1.grid(which='both')
    ax1.grid(which='minor', alpha=0.5)
    ax1.grid(which='major', alpha=0.75)
    ax1.set_xticks([x for idx, x in enumerate(thresholds) if idx % 2 == 0])
    ax1.set_xticks(thresholds, minor=True)

    # ax1.set_xlim(0, 1)
    ax1.set_xlabel('confidence threshold')

    # ax1.set_ylim(0, 1)
    # ax1.set_yticks(y_ticks_major)
    # ax1.set_yticklabels(y_ticks_major_labels)
    # ax1.set_yticks(y_ticks_minor, minor=True)

    for i in range(len(scores)):
        algorithm_name = algorithm_names[
                             i] + ' ' if algorithm_names is not None and i < len(
            algorithm_names) else ''
        ax1.plot(thresholds, scores[i], label=algorithm_name, color=colors[i],
                 linewidth=3, marker='o')

    ax1.legend(frameon=True)
    plt.tight_layout()
    plt.show()


def roc_curves(fpr_tprs, algorithm_names=None, title=None, graded_color=False):
    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab10').colors
    colormap = plt.get_cmap('RdYlGn')

    y_ticks_minor = np.linspace(0.0, 1.0, num=21)
    y_ticks_major = np.linspace(0.0, 1.0, num=11)

    fig, ax = plt.subplots()

    if title is not None:
        ax.set_title(title)

    ax.grid(which='both')
    ax.grid(which='minor', alpha=0.5)
    ax.grid(which='major', alpha=0.75)

    ax.set_xlim(0, 1)
    ax.set_xlabel('False positive rate')

    ax.set_ylim(0, 1)
    ax.set_yticks(y_ticks_major)
    ax.set_yticks(y_ticks_minor, minor=True)
    ax.set_ylabel('True positive rate')

    plt.plot([0, 1], [0, 1], color='black', linewidth=3, linestyle='--')

    for i in range(len(fpr_tprs)):
        algorithm_name = algorithm_names[
                             i] + ' ' if algorithm_names is not None and i < len(
            algorithm_names) else ''
        color = colormap(i / len(fpr_tprs)) if graded_color else colors[i]
        ax.plot(fpr_tprs[i][0], fpr_tprs[i][1], label=algorithm_name,
                color=color,
                linewidth=3)

    ax.legend(frameon=True)
    plt.tight_layout()
    plt.show()


def calibration_plot(fraction_positives, mean_predicted_values,
                     algorithm_names=None):
    assert len(fraction_positives) == len(mean_predicted_values)

    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab10').colors

    num_algorithms = len(fraction_positives)

    plt.figure(figsize=(9, 9))
    plt.grid(which='both')
    plt.grid(which='minor', alpha=0.5)
    plt.grid(which='major', alpha=0.75)

    plt.plot([0, 1], [0, 1], 'k:', label='Perfectly calibrated')

    for i in range(num_algorithms):
        # ax1.plot(mean_predicted_values[i], fraction_positives[i],
        #         label=algorithms[i] if algorithm_names is not None and i < len(algorithms) else '')

        # sns.tsplot(mean_predicted_values[i], fraction_positives[i], ax=ax1, color=colors[i])

        sns.regplot(mean_predicted_values[i], fraction_positives[i],
                    order=3, x_estimator=np.mean, color=colors[i], marker='o',
                    scatter_kws={'s': 40},
                    label=algorithm_names[
                        i] if algorithm_names is not None and i < len(
                        algorithm_names) else '')

    ticks = np.linspace(0.0, 1.0, num=11)
    plt.xlim([-0.05, 1.05])
    plt.xticks(ticks)
    plt.xlabel('Predicted probability')
    plt.ylabel('Observed probability')
    plt.ylim([-0.05, 1.05])
    plt.yticks(ticks)
    plt.legend(loc='lower right')
    plt.title('Calibration (reliability curve)')

    plt.tight_layout()
    plt.show()


def brier_plot(brier_scores, algorithm_names=None, title=None):
    sns.set_style('whitegrid')

    if title is not None:
        plt.title(title)

    colors = plt.get_cmap('tab10').colors

    plt.grid(which='both')
    plt.grid(which='minor', alpha=0.5)
    plt.grid(which='major', alpha=0.75)
    plt.xlabel('class')
    plt.ylabel('brier')

    x = np.array(range(brier_scores.shape[0]))
    for i in range(brier_scores.shape[1]):
        plt.plot(brier_scores[:, i],
                 label=algorithm_names[
                           i] + ' ' if algorithm_names is not None and i < len(
                     algorithm_names) else '',
                 color=colors[i], linewidth=3)

    plt.legend()
    plt.tight_layout()
    plt.show()


def predictions_distribution_plot(probabilities, algorithm_names=None):
    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab10').colors

    num_algorithms = len(probabilities)

    plt.figure(figsize=(9, 9))
    plt.grid(which='both')
    plt.grid(which='minor', alpha=0.5)
    plt.grid(which='major', alpha=0.75)

    for i in range(num_algorithms):
        plt.hist(probabilities[i], range=(0, 1), bins=41, color=colors[i],
                 label=algorithm_names[
                     i] if algorithm_names is not None and i < len(
                     algorithm_names) else '',
                 histtype='stepfilled', alpha=0.5, lw=2)

    plt.xlabel('Mean predicted value')
    plt.xlim([0, 1])
    plt.xticks(np.linspace(0.0, 1.0, num=21))
    plt.ylabel('Count')
    plt.legend(loc='upper center', ncol=2)

    plt.tight_layout()
    plt.show()


def confusion_matrix_plot(confusion_matrix, labels=None, field=None):
    mpl.rcParams.update({'figure.autolayout': True})
    fig, ax = plt.subplots()

    ax.invert_yaxis()
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position('top')

    cax = ax.matshow(confusion_matrix, cmap='viridis')

    ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
    ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
    ax.set_xticklabels([''] + labels, rotation=45, ha='left')
    ax.set_yticklabels([''] + labels)
    ax.grid(False)
    ax.tick_params(axis='both', which='both', length=0)
    fig.colorbar(cax, ax=ax, extend='max')
    ax.set_xlabel('Predicted {}'.format(field))
    ax.set_ylabel('Actual {}'.format(field))

    plt.tight_layout()
    plt.show()


def double_axis_line_plot(y1_sorted, y2, y1_name, y2_name, labels=None,
                          title=None):
    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab10').colors

    fig, ax1 = plt.subplots()

    if title is not None:
        ax1.set_title(title)

    # ax1.grid(which='both')
    # ax1.grid(which='minor', alpha=0.5)
    # ax1.grid(which='major', alpha=0.75)

    ax1.set_xlabel('class (sorted by {})'.format(y1_name))
    ax1.set_xlim(0, len(y1_sorted) - 1)
    if labels is not None:
        ax1.set_xticklabels(labels, rotation=45, ha='right')
        ax1.set_xticks(np.arange(len(labels)))

    ax1.set_ylabel(y2_name, color=colors[1])
    ax1.tick_params('y', colors=colors[1])
    ax1.set_ylim(min(y2), max(y2))

    ax2 = ax1.twinx()
    ax2.set_ylabel(y1_name, color=colors[0])
    ax2.tick_params('y', colors=colors[0])
    ax2.set_ylim(min(y1_sorted), max(y1_sorted))

    ax1.plot(y2, label=y2_name, color=colors[1],
             linewidth=3)
    ax2.plot(y1_sorted, label=y1_name, color=colors[0],
             linewidth=4.0)

    fig.tight_layout()
    plt.show()


def plot_matrix(matrix, cmap='hot'):
    plt.matshow(matrix, cmap=cmap)
    plt.show()


def plot_distributions(distributions, labels=None, title=None):
    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab10').colors

    fig, ax1 = plt.subplots()

    if title is not None:
        ax1.set_title(title)

    ax1.grid(which='both')
    ax1.grid(which='minor', alpha=0.5)
    ax1.grid(which='major', alpha=0.75)

    ax1.set_xlabel('class')

    ax1.set_ylabel('p')
    ax1.tick_params('y')

    for i, distribution in enumerate(distributions):
        ax1.plot(distribution, color=colors[i], alpha=0.6,
                 label=labels[i] if labels is not None and i < len(
                     labels) else 'Distribution {}'.format(i))

    ax1.legend(frameon=True)
    fig.tight_layout()
    plt.show()


def plot_distributions_difference(distribution, labels=None, title=None):
    sns.set_style('whitegrid')

    colors = plt.get_cmap('tab10').colors

    fig, ax1 = plt.subplots()

    if title is not None:
        ax1.set_title(title)

    ax1.grid(which='both')
    ax1.grid(which='minor', alpha=0.5)
    ax1.grid(which='major', alpha=0.75)

    ax1.set_xlabel('class')

    ax1.set_ylabel('p')
    ax1.tick_params('y')

    ax1.plot(distribution, color=colors[0])

    fig.tight_layout()
    plt.show()


def bar_plot(xs, ys, decimals=4, labels=None, title=None):
    assert len(xs) == len(ys)
    assert len(xs) > 0

    sns.set_style('whitegrid')

    fig, ax = plt.subplots()

    ax.grid(which='both')
    ax.grid(which='minor', alpha=0.5)
    ax.grid(which='major', alpha=0.75)

    if title is not None:
        ax.set_title(title)

    colors = plt.get_cmap('tab10').colors

    ax.invert_yaxis()  # labels read top-to-bottom

    maximum = ys.max()
    ticks = np.arange(len(xs))
    ax.set_yticks(ticks)
    if labels is None:
        ax.set_yticklabels(xs)
    else:
        ax.set_yticklabels(labels)

    ax.barh(ticks, ys, color=colors[0], align='center')

    for i, v in enumerate(ys):
        if v < maximum * (0.025 * decimals + 0.1):
            x = v + maximum * 0.01
            horizontal_alignment = 'left'
        else:
            x = v - maximum * 0.01
            horizontal_alignment = 'right'
        txt = ax.text(x, ticks[i], ('{:.' + str(decimals) + 'f}').format(v),
                      color='white',
                      fontweight='bold', verticalalignment='center',
                      horizontalalignment=horizontal_alignment)
        txt.set_path_effects(
            [PathEffects.withStroke(linewidth=3, foreground='black')])

    plt.tight_layout()
    plt.show()



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging

nlp_pipeline = None
default_characters = [' ', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
                      'k', 'l', 'm', 'n', 'o', 'p',
                      'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0',
                      '1', '2', '3', '4', '5', '6',
                      '8', '9', '-', ',', ';', '.', '!', '?', ':', '\'', '\'',
                      '/', '\\', '|', '_', '@', '#',
                      '$', '%', '^', '&', '*', '~', '`', '+', '-', '=', '<',
                      '>', '(', ')', '[', ']', '{',
                      '}']
punctuation = {'.', ',', '@', '$', '%', '/', ':', ';', '+', '='}


def load_nlp_pipeline():
    global nlp_pipeline
    if nlp_pipeline is None:
        logging.info('Loading NLP pipeline')
        try:
            import en_core_web_sm
        except FileNotFoundError:
            logging.error('Unable to load spacy model en_core_web_sm. '
                          'Make sure to download it with: '
                          'python -m spacy download en')
        nlp_pipeline = en_core_web_sm.load(disable=['parser', 'tagger', 'ner'])
    return nlp_pipeline


def pass_filters(
        token,
        filter_numbers=False,
        filter_punctuation=False,
        filter_short_tokens=False,
        filter_stopwords=False
):
    passes_filters = True
    if filter_numbers:
        passes_filters = not token.like_num
    if passes_filters and filter_punctuation:
        passes_filters = not bool(set(token.orth_) & punctuation)
    if passes_filters and filter_short_tokens:
        passes_filters = len(token) > 2
    if passes_filters and filter_stopwords:
        passes_filters = not token.is_stop
    return passes_filters


def process_text(
        text,
        nlp_pipeline,
        return_lemma=False,
        filter_numbers=False,
        filter_punctuation=False,
        filter_short_tokens=False,
        filter_stopwords=False
):
    doc = nlp_pipeline.tokenizer(text)
    return [token.lemma_ if return_lemma else token.text
            for token in doc if pass_filters(token,
                                             filter_numbers,
                                             filter_punctuation,
                                             filter_short_tokens,
                                             filter_stopwords)
            ]


# class Lemmatizer(object):
#     def __init__(self):
#         self.pipeline = load_nlp_pipeline()
#
#     def __call__(self, doc):
#         tokens = [t.lemma_ for t in self.pipeline(doc)
#                   if not t.like_num and len(t) > 2 and '.' not in t.orth_ and '/' not in t.orth_]
#         return tokens
#
#
# class Vectorizer():
#     def __init__(self, weighting, min_df=0.005, max_df=0.995):
#         self.tvectorizer = TfidfVectorizer(analyzer='word',
#                                            min_df=min_df,
#                                            max_df=max_df,
#                                            tokenizer=Lemmatizer(),
#                                            ngram_range=(1, 1))
#
#     def fit_transform(self, docs):
#         return self.tvectorizer.fit_transform(docs).toarray()
#
#     def transform(self, docs):
#         return self.tvectorizer.transform(docs).toarray()

if __name__ == '__main__':
    text = 'Hello John, how are you doing my good old friend? Are you still number 732 in the list? Did you pay $32.43 or 54.21 for the book?'
    print(process_text(text, load_nlp_pipeline()))
    print(process_text(text, load_nlp_pipeline(),
                       filter_numbers=True,
                       filter_punctuation=True,
                       filter_short_tokens=True))
    print(process_text(text, load_nlp_pipeline(),
                       filter_stopwords=True))
    print(process_text(text, load_nlp_pipeline(),
                       return_lemma=True))
    print(process_text(text, load_nlp_pipeline(),
                       return_lemma=True,
                       filter_numbers=True,
                       filter_punctuation=True,
                       filter_short_tokens=True))
    print(process_text(text, load_nlp_pipeline(),
                       return_lemma=True,
                       filter_stopwords=True))
